# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import collections
import datetime
import unittest

from cryptodatahub.common.exception import InvalidValue
from cryptodatahub.common.types import Base64Data
from cryptodatahub.common.stores import CertificateTransparencyLog
from cryptodatahub.tls.algorithm import (
    TlsECPointFormat,
    TlsCertificateCompressionAlgorithm,
    TlsNamedCurve,
    TlsNextProtocolName,
    TlsProtocolName,
    TlsPskKeyExchangeMode,
    TlsSignatureAndHashAlgorithm,
    TlsTokenBindingParamater,
)

from cryptoparser.common.exception import NotEnoughData, InvalidType
from cryptoparser.common.x509 import CtExtensions, CtVersion, SignedCertificateTimestamp

from cryptoparser.tls.extension import (
    TlsCertificateStatusRequestExtensions,
    TlsCertificateStatusRequestResponderId,
    TlsCertificateStatusRequestResponderIdList,
    TlsExtensionApplicationLayerProtocolNegotiation,
    TlsExtensionApplicationLayerProtocolSettings,
    TlsExtensionCertificateStatusRequestClient,
    TlsExtensionCertificateStatusRequestServer,
    TlsExtensionChannelId,
    TlsExtensionCompressCertificate,
    TlsExtensionECPointFormats,
    TlsExtensionEllipticCurves,
    TlsExtensionEncryptedClientHelloInner,
    TlsExtensionEncryptedClientHelloOuter,
    TlsExtensionEncryptThenMAC,
    TlsExtensionExtendedMasterSecret,
    TlsExtensionKeyShareClient,
    TlsExtensionKeyShareClientHelloRetry,
    TlsExtensionKeyShareServer,
    TlsExtensionKeyShareReservedClient,
    TlsExtensionNextProtocolNegotiationClient,
    TlsExtensionNextProtocolNegotiationServer,
    TlsExtensionPadding,
    TlsExtensionPostHandshakeAuthentication,
    TlsExtensionPskKeyExchangeModes,
    TlsExtensionRecordSizeLimit,
    TlsExtensionRenegotiationInfo,
    TlsExtensionServerNameClient,
    TlsExtensionServerNameServer,
    TlsExtensionSessionTicket,
    TlsExtensionShortRecordHeader,
    TlsExtensionSignatureAlgorithms,
    TlsExtensionSignatureAlgorithmsCert,
    TlsExtensionSignedCertificateTimestampClient,
    TlsExtensionSignedCertificateTimestampServer,
    TlsExtensionSupportedVersionsClient,
    TlsExtensionSupportedVersionsServer,
    TlsExtensionTokenBinding,
    TlsExtensionUnparsed,
    TlsExtensionParsed,
    TlsExtensionType,
    TlsNextProtocolNameList,
    TlsProtocolNameList,
    TlsRenegotiatedConnection,
    TlsTokenBindingProtocolVersion,
)
from cryptoparser.tls.grease import TlsGreaseOneByte, TlsGreaseTwoByte, TlsInvalidTypeOneByte, TlsInvalidTypeTwoByte
from cryptoparser.tls.version import TlsVersion, TlsProtocolVersion

from .classes import TestUnusedDataExtension


class TestExtensionUnparsed(unittest.TestCase):

    def test_error(self):
        extension_missing_data_dict = collections.OrderedDict([
            ('extension_type', b'\xff\x01'),
            ('extension_length', b'\x00\x01'),
        ])
        extension_missing_data_bytes = b''.join(extension_missing_data_dict.values())
        with self.assertRaises(NotEnoughData) as context_manager:
            # pylint: disable=expression-not-assigned
            TlsExtensionUnparsed.parse_exact_size(extension_missing_data_bytes)
        self.assertEqual(context_manager.exception.bytes_needed, 5)

    def test_parse_and_compose(self):
        extension_minimal_dict = collections.OrderedDict([
            ('extension_type', b'\xff\x01'),
            ('extension_length', b'\x00\x00'),
        ])
        extension_minimal_bytes = b''.join(extension_minimal_dict.values())

        extension_minimal = TlsExtensionUnparsed.parse_exact_size(extension_minimal_bytes)
        self.assertEqual(extension_minimal.compose(), extension_minimal_bytes)

        extension_with_data_dict = collections.OrderedDict([
            ('extension_type', b'\xff\x01'),
            ('extension_length', b'\x00\x04'),
            ('extension_data', b'\xde\xad\xbe\xaf'),
        ])
        extension_with_data_bytes = b''.join(extension_with_data_dict.values())

        extension_with_data = TlsExtensionUnparsed.parse_exact_size(extension_with_data_bytes)
        self.assertEqual(extension_with_data.compose(), extension_with_data_bytes)


class ExtensionInvalidType(TlsExtensionParsed):
    @classmethod
    def get_extension_type(cls):
        return 0xffff

    @classmethod
    def _parse(cls, parsable):
        parser = super(ExtensionInvalidType, cls)._parse_header(parsable)

        return ExtensionInvalidType(), parser.parsed_length

    def compose(self):
        raise NotImplementedError


class TestExtensionParsed(unittest.TestCase):
    def test_error(self):
        extension_invalid_type_dict = collections.OrderedDict([
            ('extension_type', b'\x00\x00'),
            ('extension_length', b'\x00\x00'),
        ])
        extension_invalid_type_bytes = b''.join(extension_invalid_type_dict.values())

        with self.assertRaises(InvalidType):
            # pylint: disable=expression-not-assigned
            ExtensionInvalidType.parse_exact_size(extension_invalid_type_bytes)


class TestExtensionHostnameClient(unittest.TestCase):
    def test_parse(self):
        extension_hostname_dict = collections.OrderedDict([
            ('extension_type', b'\x00\x00'),
            ('extension_length', b'\x00\x14'),
            ('name_list_length', b'\x00\x12'),
            ('name_type', b'\x00'),
            ('name_length', b'\x00\x0f'),
            ('name', b'\x77\x77\x77\x2e\x65\x78\x61\x6d\x70\x6c\x65\x2e\x63\x6f\x6d')
        ])
        extension_hostname_bytes = b''.join(extension_hostname_dict.values())

        extension_hostname = TlsExtensionServerNameClient.parse_exact_size(extension_hostname_bytes)
        self.assertEqual(extension_hostname.host_name, 'www.example.com')
        self.assertEqual(extension_hostname.compose(), extension_hostname_bytes)

        extension_hostname_internationalized_dict = collections.OrderedDict([
            ('extension_type', b'\x00\x00'),
            ('extension_length', b'\x00\x1e'),
            ('name_list_length', b'\x00\x1c'),
            ('name_type', b'\x00'),
            ('name_length', b'\x00\x19'),
            (
                'name',
                b'\x78\x6e\x2d\x2d\x73\x6c\x61\x6e\x64\x2d\x79\x73\x61\x2e\x69\x63\x6f\x6d\x2e\x6d\x75\x73\x65\x75\x6d'
            )
        ])
        extension_hostname_internationalized_bytes = b''.join(extension_hostname_internationalized_dict.values())

        extension_hostname_internationalized = TlsExtensionServerNameClient.parse_exact_size(
            extension_hostname_internationalized_bytes
        )
        self.assertEqual(extension_hostname_internationalized.host_name, 'ísland.icom.museum')
        self.assertEqual(extension_hostname_internationalized.compose(), extension_hostname_internationalized_bytes)


class TestExtensionECPointFormat(unittest.TestCase):
    def test_parse(self):
        extension_ec_point_formats_dict = collections.OrderedDict([
            ('extension_type', b'\x00\x0b'),
            ('extension_length', b'\x00\x03'),
            ('ec_point_format_list_length', b'\x02'),
            ('ec_point_format_list', b'\x00\x0b'),
        ])
        extension_ec_point_formats_bytes = b''.join(extension_ec_point_formats_dict.values())

        extension_point_formats = TlsExtensionECPointFormats.parse_exact_size(extension_ec_point_formats_bytes)
        self.assertEqual(
            list(extension_point_formats.point_formats),
            [
                TlsECPointFormat.UNCOMPRESSED,
                TlsInvalidTypeOneByte(TlsGreaseOneByte.GREASE_0B),
            ]
        )
        self.assertEqual(extension_point_formats.compose(), extension_ec_point_formats_bytes)


class TestExtensionHostnameServer(unittest.TestCase):
    def test_parse(self):
        extension_hostname_dict = collections.OrderedDict([
            ('extension_type', b'\x00\x00'),
            ('extension_length', b'\x00\x00'),
        ])
        extension_hostname_bytes = b''.join(extension_hostname_dict.values())

        extension_hostname = TlsExtensionServerNameServer.parse_exact_size(extension_hostname_bytes)
        self.assertEqual(extension_hostname.compose(), extension_hostname_bytes)


class TestExtensionEllipticCurves(unittest.TestCase):
    def test_parse(self):
        extension_elliptic_curves_dict = collections.OrderedDict([
            ('extension_type', b'\x00\x0a'),
            ('extension_length', b'\x00\x0a'),
            ('elliptic_curve_list_length', b'\x00\x08'),
            ('elliptic_curve_list', b'\x00\x1d\x00\x17\x00\x18\x0a\x0a'),
        ])
        extension_elliptic_curves_bytes = b''.join(extension_elliptic_curves_dict.values())

        extension_elliptic_curves = TlsExtensionEllipticCurves.parse_exact_size(extension_elliptic_curves_bytes)
        self.assertEqual(
            list(extension_elliptic_curves.elliptic_curves),
            [
                TlsNamedCurve.X25519,
                TlsNamedCurve.SECP256R1,
                TlsNamedCurve.SECP384R1,
                TlsInvalidTypeTwoByte(TlsGreaseTwoByte.GREASE_0A0A),
            ]
        )
        self.assertEqual(extension_elliptic_curves.compose(), extension_elliptic_curves_bytes)


class TestExtensionSupportedVersions(unittest.TestCase):
    def test_parse(self):
        extension_supported_versions_dict = collections.OrderedDict([
            ('extension_type', b'\x00\x2b'),
            ('extension_length', b'\x00\x09'),
            ('supported_version_list_length', b'\x08'),
            ('supported_version_list', b'\x03\x02\x03\x03\x7f\x18\x0a\x0a'),
        ])
        extension_supported_versions_bytes = b''.join(extension_supported_versions_dict.values())
        extension_supported_versions = TlsExtensionSupportedVersionsClient.parse_exact_size(
            extension_supported_versions_bytes
        )
        self.assertEqual(
            list(extension_supported_versions.supported_versions),
            [
                TlsProtocolVersion(TlsVersion.TLS1_1),
                TlsProtocolVersion(TlsVersion.TLS1_2),
                TlsProtocolVersion(TlsVersion.TLS1_3_DRAFT_24),
                TlsInvalidTypeTwoByte(TlsGreaseTwoByte.GREASE_0A0A.value.code),
            ]
        )
        self.assertEqual(extension_supported_versions.compose(), extension_supported_versions_bytes)

        extension_supported_versions_dict = collections.OrderedDict([
            ('extension_type', b'\x00\x2b'),
            ('extension_length', b'\x00\x02'),
            ('selected_version', b'\x03\x03'),
        ])
        extension_supported_versions_bytes = b''.join(extension_supported_versions_dict.values())
        extension_supported_versions = TlsExtensionSupportedVersionsServer.parse_exact_size(
            extension_supported_versions_bytes
        )
        self.assertEqual(
            extension_supported_versions.selected_version,
            TlsProtocolVersion(TlsVersion.TLS1_2)
        )
        self.assertEqual(extension_supported_versions.compose(), extension_supported_versions_bytes)


class TestExtensionTokenBinding(unittest.TestCase):
    def test_parse(self):
        extension_token_binding_dict = collections.OrderedDict([
            ('extension_type', b'\x00\x18'),
            ('extension_length', b'\x00\x06'),
            ('protocol_version', b'\x01\x02'),
            ('supported_version_list', b'\x03\x02\x01\x00'),
        ])
        extension_token_binding_bytes = b''.join(extension_token_binding_dict.values())
        extension_token_binding = TlsExtensionTokenBinding.parse_exact_size(
            extension_token_binding_bytes
        )
        self.assertEqual(
            list(extension_token_binding.parameters),
            [
                TlsTokenBindingParamater.ECDSAP256,
                TlsTokenBindingParamater.RSA2048_PSS,
                TlsTokenBindingParamater.RSA2048_PKCS1_5,
            ]
        )
        self.assertEqual(
            extension_token_binding.protocol_version,
            TlsTokenBindingProtocolVersion(1, 2),
        )
        self.assertEqual(extension_token_binding.compose(), extension_token_binding_bytes)


class TestExtensionCompressCertificateAlgorithms(unittest.TestCase):
    def test_parse(self):
        extension_compress_certificate_dict = collections.OrderedDict([
            ('extension_type', b'\x00\x1b'),
            ('extension_length', b'\x00\x07'),
            ('compression_algorithm_list_length', b'\x06'),
            ('compression_algorithm_list', b'\x00\x03\x00\x02\x00\x01'),
        ])
        extension_compress_certificate_bytes = b''.join(extension_compress_certificate_dict.values())
        extension_compress_certificate = TlsExtensionCompressCertificate.parse_exact_size(
            extension_compress_certificate_bytes
        )
        self.assertEqual(extension_compress_certificate.extension_type, TlsExtensionType.COMPRESS_CERTIFICATE)
        self.assertEqual(
            list(extension_compress_certificate.compression_algorithms),
            [
                TlsCertificateCompressionAlgorithm.ZSTD,
                TlsCertificateCompressionAlgorithm.BROTLI,
                TlsCertificateCompressionAlgorithm.ZLIB,
            ]
        )
        self.assertEqual(extension_compress_certificate.compose(), extension_compress_certificate_bytes)


class TestExtensionSignatureAlgorithms(unittest.TestCase):
    def test_parse(self):
        extension_signature_algorithms_dict = collections.OrderedDict([
            ('extension_type', b'\x00\x0d'),
            ('extension_length', b'\x00\x0c'),
            ('signature_algorithm_list_length', b'\x00\x0a'),
            ('signature_algorithm_list', b'\x01\x00\x02\x01\x03\x02\x04\x03\x0a\x0a'),
        ])
        extension_signature_algorithms_bytes = b''.join(extension_signature_algorithms_dict.values())
        extension_signature_algorithms = TlsExtensionSignatureAlgorithms.parse_exact_size(
            extension_signature_algorithms_bytes
        )
        self.assertEqual(extension_signature_algorithms.extension_type, TlsExtensionType.SIGNATURE_ALGORITHMS)
        self.assertEqual(
            list(extension_signature_algorithms.hash_and_signature_algorithms),
            [
                TlsSignatureAndHashAlgorithm.ANONYMOUS_MD5,
                TlsSignatureAndHashAlgorithm.RSA_SHA1,
                TlsSignatureAndHashAlgorithm.DSA_SHA224,
                TlsSignatureAndHashAlgorithm.ECDSA_SHA256,
                TlsInvalidTypeTwoByte(TlsGreaseTwoByte.GREASE_0A0A),
            ]
        )
        self.assertEqual(extension_signature_algorithms.compose(), extension_signature_algorithms_bytes)


class TestExtensionSignatureAlgorithmsCert(unittest.TestCase):
    def test_parse(self):
        extension_signature_algorithms_cert_dict = collections.OrderedDict([
            ('extension_type', b'\x00\x32'),
            ('extension_length', b'\x00\x0a'),
            ('signature_algorithm_list_length', b'\x00\x08'),
            ('signature_algorithm_list', b'\x01\x00\x02\x01\x03\x02\x04\x03'),
        ])
        extension_signature_algorithms_cert_bytes = b''.join(extension_signature_algorithms_cert_dict.values())
        extension_signature_algorithms_cert = TlsExtensionSignatureAlgorithmsCert.parse_exact_size(
            extension_signature_algorithms_cert_bytes
        )
        self.assertEqual(extension_signature_algorithms_cert.extension_type, TlsExtensionType.SIGNATURE_ALGORITHMS_CERT)
        self.assertEqual(
            list(extension_signature_algorithms_cert.hash_and_signature_algorithms),
            [
                TlsSignatureAndHashAlgorithm.ANONYMOUS_MD5,
                TlsSignatureAndHashAlgorithm.RSA_SHA1,
                TlsSignatureAndHashAlgorithm.DSA_SHA224,
                TlsSignatureAndHashAlgorithm.ECDSA_SHA256,
            ]
        )
        self.assertEqual(extension_signature_algorithms_cert.compose(), extension_signature_algorithms_cert_bytes)


class TestExtensionSignedCertificateTimestampClient(unittest.TestCase):
    def test_parse_minimal(self):
        extension_sct_dict = collections.OrderedDict([
            ('extension_type', b'\x00\x12'),
            ('extension_length', b'\x00\x00'),
        ])
        extension_sct_bytes = b''.join(extension_sct_dict.values())
        extension_sct = TlsExtensionSignedCertificateTimestampClient.parse_exact_size(
            extension_sct_bytes
        )
        self.assertEqual(
            extension_sct.extension_type, TlsExtensionType.SIGNED_CERTIFICATE_TIMESTAMP
        )
        self.assertEqual(extension_sct.compose(), extension_sct_bytes)


class TestExtensionSignedCertificateTimestampServer(unittest.TestCase):
    def setUp(self):
        extension_sct_dict = collections.OrderedDict([
            ('extension_type', b'\x00\x12'),
            ('extension_length', b'\x00\x84'),
            ('signed_certificate_timestamp_list_length', b'\x00\x82'),
            ('signed_certificate_timestamp_list', b''.join([
                b'\x00\x3f',                                                                  # length
                b'\x00',                                                                      # version
                b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f',
                b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f',          # log
                b'\x00\x00\x00\x00\x00\x00\x00\x01',                                          # timestamp
                b'\x00\x00',                                                                  # extensions
                b'\x00\x00',                                                                  # signature_algorithm
                b'\x00\x10\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f',  # signature

                b'\x00\x3f',                                                                  # length
                b'\x00',                                                                      # version
                b'\x96\x06\xc0\x2c\x69\x00\x33\xaa\x1d\x14\x5f\x59\xc6\xe2\x64\x8d',
                b'\x05\x49\xf0\xdf\x96\xaa\xb8\xdb\x91\x5a\x70\xd8\xec\xf3\x90\xa5',          # log
                b'\x00\x00\x00\x00\x00\x00\x00\x01',                                          # timestamp
                b'\x00\x00',                                                                  # extensions
                b'\x00\x00',                                                                  # signature_algorithm
                b'\x00\x10\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f',  # signature
            ]))
        ])
        self.extension_sct_bytes = b''.join(extension_sct_dict.values())

    def test_parse(self):
        extension_sct = TlsExtensionSignedCertificateTimestampServer.parse_exact_size(
            self.extension_sct_bytes
        )
        self.assertEqual(extension_sct.extension_type, TlsExtensionType.SIGNED_CERTIFICATE_TIMESTAMP)

        sct = SignedCertificateTimestamp(
            CtVersion.V1,
            Base64Data(2 * b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'),
            datetime.datetime(1970, 1, 1, 0, 0, 0, 1000, tzinfo=datetime.timezone.utc),
            CtExtensions([]),
            TlsSignatureAndHashAlgorithm.ANONYMOUS_NONE,
            b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f',
        )
        self.assertEqual(extension_sct.scts[0], sct)

        sct.log = CertificateTransparencyLog.AKAMAI_CT_LOG.value
        self.assertEqual(extension_sct.scts[1], sct)

        self.assertEqual(extension_sct.compose(), self.extension_sct_bytes)

    def test_markdown(self):
        scts = TlsExtensionSignedCertificateTimestampServer.parse_exact_size(
            self.extension_sct_bytes
        ).scts

        self.assertEqual(scts[0].as_markdown(), '\n'.join([
            '* Version: V1',
            '* Log: AAECAwQFBgcICQoLDA0ODwABAgMEBQYHCAkKCwwNDg8=',
            '* Timestamp: 1970-01-01 00:00:00.001000+00:00',
            '* Extensions: -',
            '* Signature Algorithm: none with no encryption',
            ''
        ]))

        self.assertEqual(scts[1].as_markdown(), '\n'.join([
            '* Version: V1',
            '* Log: Akamai CT Log (lgbALGkAM6odFF9ZxuJkjQVJ8N+WqrjbkVpw2OzzkKU=)',
            '* Timestamp: 1970-01-01 00:00:00.001000+00:00',
            '* Extensions: -',
            '* Signature Algorithm: none with no encryption',
            ''
        ]))


class TestExtensionKeyShareClient(unittest.TestCase):
    def test_parse(self):
        extension_key_share_dict = collections.OrderedDict([
            ('extension_type', b'\x00\x33'),
            ('extension_length', b'\x00\x2a'),
            ('key_share_length', b'\x00\x28'),
            ('group_grease', b'\x0a\x0a'),
            ('key_exchange_length_grease', b'\x00\x00'),
            ('group', b'\x00\x1d'),
            ('key_exchange_length', b'\x00\x20'),
            ('key_exchange',
             b'\x00\x01\x02\x03\x04\x05\x06\x07' +
             b'\x00\x01\x02\x03\x04\x05\x06\x07' +
             b'\x00\x01\x02\x03\x04\x05\x06\x07' +
             b'\x00\x01\x02\x03\x04\x05\x06\x07' +
             b'')
        ])
        extension_key_share_bytes = b''.join(extension_key_share_dict.values())
        extension_key_share = TlsExtensionKeyShareClient.parse_exact_size(
            extension_key_share_bytes
        )
        key_share_entries = extension_key_share.key_share_entries
        self.assertEqual(len(key_share_entries), 2)
        self.assertEqual(key_share_entries[0].group, TlsInvalidTypeTwoByte(TlsGreaseTwoByte.GREASE_0A0A))
        self.assertEqual(key_share_entries[1].group, TlsNamedCurve.X25519)
        self.assertEqual(
            bytearray(key_share_entries[1].key_exchange),
            extension_key_share_dict['key_exchange']
        )
        self.assertEqual(extension_key_share.compose(), extension_key_share_bytes)


class TestExtensionKeyShareReservedClient(unittest.TestCase):
    def test_parse(self):
        extension_key_share_dict = collections.OrderedDict([
            ('extension_type', b'\x00\x28'),
            ('extension_length', b'\x00\x26'),
            ('key_share_length', b'\x00\x24'),
            ('group', b'\x00\x1d'),
            ('key_exchange_length', b'\x00\x20'),
            ('key_exchange',
             b'\x00\x01\x02\x03\x04\x05\x06\x07' +
             b'\x00\x01\x02\x03\x04\x05\x06\x07' +
             b'\x00\x01\x02\x03\x04\x05\x06\x07' +
             b'\x00\x01\x02\x03\x04\x05\x06\x07' +
             b'')
        ])
        extension_key_share_bytes = b''.join(extension_key_share_dict.values())
        extension_key_share = TlsExtensionKeyShareReservedClient.parse_exact_size(
            extension_key_share_bytes
        )
        key_share_entries = extension_key_share.key_share_entries  # pylint: disable=no-member
        self.assertEqual(len(key_share_entries), 1)
        self.assertEqual(key_share_entries[0].group, TlsNamedCurve.X25519)
        self.assertEqual(
            bytearray(key_share_entries[0].key_exchange),
            extension_key_share_dict['key_exchange']
        )
        self.assertEqual(extension_key_share.compose(), extension_key_share_bytes)


class TestExtensionKeyShareClientHelloRetry(unittest.TestCase):
    def test_error(self):
        extension_key_share_dict = collections.OrderedDict([
            ('extension_type', b'\x00\x33'),
            ('extension_length', b'\x00\x04'),
            ('data', b'\x00\x00\x00\x00'),
        ])
        extension_key_share_bytes = b''.join(extension_key_share_dict.values())

        with self.assertRaises(InvalidType):
            # pylint: disable=expression-not-assigned
            TlsExtensionKeyShareClientHelloRetry.parse_exact_size(
                extension_key_share_bytes
            )

    def test_parse(self):
        extension_key_share_dict = collections.OrderedDict([
            ('extension_type', b'\x00\x33'),
            ('extension_length', b'\x00\x02'),
            ('group', b'\x00\x1d'),
        ])
        extension_key_share_bytes = b''.join(extension_key_share_dict.values())
        extension_key_share = TlsExtensionKeyShareClientHelloRetry.parse_exact_size(
            extension_key_share_bytes
        )
        self.assertEqual(extension_key_share.selected_group, TlsNamedCurve.X25519)
        self.assertEqual(extension_key_share.compose(), extension_key_share_bytes)


class TestExtensionKeyShareServer(unittest.TestCase):
    def test_parse(self):
        extension_key_share_dict = collections.OrderedDict([
            ('extension_type', b'\x00\x33'),
            ('extension_length', b'\x00\x24'),
            ('group', b'\x00\x1d'),
            ('key_exchange_length', b'\x00\x20'),
            ('key_exchange',
             b'\x00\x01\x02\x03\x04\x05\x06\x07' +
             b'\x00\x01\x02\x03\x04\x05\x06\x07' +
             b'\x00\x01\x02\x03\x04\x05\x06\x07' +
             b'\x00\x01\x02\x03\x04\x05\x06\x07' +
             b'')
        ])
        extension_key_share_bytes = b''.join(extension_key_share_dict.values())
        extension_key_share = TlsExtensionKeyShareServer.parse_exact_size(
            extension_key_share_bytes
        )
        self.assertEqual(extension_key_share.key_share_entry.group, TlsNamedCurve.X25519)
        self.assertEqual(
            bytearray(extension_key_share.key_share_entry.key_exchange),
            extension_key_share_dict['key_exchange']
        )
        self.assertEqual(extension_key_share.compose(), extension_key_share_bytes)


class TestExtensionCertificateStatusRequestClient(unittest.TestCase):
    def setUp(self):
        self.status_request_minimal_bytes = bytes(
            b'\x00\x05' +                          # handshake_type = STATUS_REQUEST
            b'\x00\x05' +                          # length = 0x05
            b'\x01' +                              # status_type = OCSP
            b'\x00\x00' +                          # responder_id_list_length = 0x00
            b'\x00\x00' +                          # request_extensions_length = 0x00
            b''
        )
        self.status_request_minimal = TlsExtensionCertificateStatusRequestClient()

        self.request_extensions = b'\x00\x01\x02\x03\x04\x05\x06\x07'
        self.status_request_bytes = bytes(
            b'\x00\x05' +                          # handshake_type = STATUS_REQUEST
            b'\x00\x15' +                          # length = 0x05
            b'\x01' +                              # status_type = OCSP
            b'\x00\x08' +                          # responder_id_list_length = 0x08
            b'\x00\x01\x00\x00\x03\x01\x02\x03' +  # responder_id_list
            b'\x00\x08' +                          # request_extensions_length = 0x08
            self.request_extensions +              # request_extensions
            b''
        )
        self.status_request = TlsExtensionCertificateStatusRequestClient(
            responder_id_list=TlsCertificateStatusRequestResponderIdList([
                TlsCertificateStatusRequestResponderId(b'\x00'),
                TlsCertificateStatusRequestResponderId(b'\x01\x02\x03')
            ]),
            extensions=TlsCertificateStatusRequestExtensions(self.request_extensions)
        )

    def test_parse(self):
        status_request_minimal = TlsExtensionCertificateStatusRequestClient.parse_exact_size(
            self.status_request_minimal_bytes
        )
        self.assertEqual(status_request_minimal.responder_id_list, TlsCertificateStatusRequestResponderIdList([]))
        self.assertEqual(status_request_minimal.request_extensions, TlsCertificateStatusRequestExtensions([]))

        status_request = TlsExtensionCertificateStatusRequestClient.parse_exact_size(self.status_request_bytes)
        self.assertEqual(
            status_request.responder_id_list,
            TlsCertificateStatusRequestResponderIdList([
                TlsCertificateStatusRequestResponderId(b'\x00'),
                TlsCertificateStatusRequestResponderId(b'\x01\x02\x03')
            ])
        )
        self.assertEqual(
            status_request.request_extensions,
            TlsCertificateStatusRequestExtensions(self.request_extensions)
        )

    def test_compose(self):
        self.assertEqual(self.status_request_minimal.compose(), self.status_request_minimal_bytes)
        self.assertEqual(self.status_request.compose(), self.status_request_bytes)


class TestExtensionCertificateStatusRequestServer(unittest.TestCase):
    def setUp(self):
        self.status_request_empty_bytes = bytes(
            b'\x00\x05' +                          # handshake_type = STATUS_REQUEST
            b'\x00\x00' +                          # length = 0x05
            b''
        )
        self.status_request_empty = TlsExtensionCertificateStatusRequestServer()

    def test_parse(self):
        status_request_empty = TlsExtensionCertificateStatusRequestServer.parse_exact_size(
            self.status_request_empty_bytes
        )
        self.assertEqual(status_request_empty, self.status_request_empty)

    def test_compose(self):
        self.assertEqual(self.status_request_empty.compose(), self.status_request_empty_bytes)


class TestExtensionChannelId(unittest.TestCase):
    def test_parse(self):
        extension_channel_id_dict = collections.OrderedDict([
            ('extension_type', b'\x75\x50'),
            ('extension_length', b'\x00\x00'),
        ])
        extension_channel_id_bytes = b''.join(extension_channel_id_dict.values())
        extension_channel_id = TlsExtensionChannelId.parse_exact_size(
            extension_channel_id_bytes
        )
        self.assertEqual(extension_channel_id.compose(), extension_channel_id_bytes)


class TestTlsExtensionPskKeyExchangeModes(unittest.TestCase):
    def test_parse(self):
        extension_pks_key_exchange_modes_dict = collections.OrderedDict([
            ('extension_type', b'\x00\x2d'),
            ('extension_length', b'\x00\x03'),
            ('supported_version_list', b'\x02\x01\x00'),
        ])
        extension_pks_key_exchange_modes_bytes = b''.join(extension_pks_key_exchange_modes_dict.values())
        extension_pks_key_exchange_modes = TlsExtensionPskKeyExchangeModes.parse_exact_size(
            extension_pks_key_exchange_modes_bytes
        )
        self.assertEqual(
            list(extension_pks_key_exchange_modes.key_exchange_modes),
            [
                TlsPskKeyExchangeMode.PSK_DH_KE,
                TlsPskKeyExchangeMode.PSK_KE,
            ]
        )
        self.assertEqual(extension_pks_key_exchange_modes.compose(), extension_pks_key_exchange_modes_bytes)


class TestExtensionRenegotiationInfo(unittest.TestCase):
    def test_parse(self):
        extension_renegotiation_info_dict = collections.OrderedDict([
            ('extension_type', b'\xff\x01'),
            ('extension_length', b'\x00\x09'),
            ('renegotiated_connection_length', b'\x08'),
            ('renegotiated_connection', b'\x00\x01\x02\x03\x04\x05\x06\x07'),
        ])
        extension_renegotiation_info_bytes = b''.join(extension_renegotiation_info_dict.values())
        extension_renegotiation_info = TlsExtensionRenegotiationInfo.parse_exact_size(
            extension_renegotiation_info_bytes
        )
        self.assertEqual(
            extension_renegotiation_info.renegotiated_connection,
            TlsRenegotiatedConnection(b'\x00\x01\x02\x03\x04\x05\x06\x07')
        )
        self.assertEqual(extension_renegotiation_info.compose(), extension_renegotiation_info_bytes)


class TestExtensionSessionTicket(unittest.TestCase):
    def test_parse(self):
        extension_session_ticket_dict = collections.OrderedDict([
            ('extension_type', b'\x00\x23'),
            ('extension_length', b'\x00\x08'),
            ('session_ticket', b'\x00\x01\x02\x03\x04\x05\x06\x07'),
        ])
        extension_session_ticket_bytes = b''.join(extension_session_ticket_dict.values())
        extension_session_ticket = TlsExtensionSessionTicket.parse_exact_size(
            extension_session_ticket_bytes
        )
        self.assertEqual(extension_session_ticket.session_ticket, b'\x00\x01\x02\x03\x04\x05\x06\x07')
        self.assertEqual(extension_session_ticket.compose(), extension_session_ticket_bytes)


class TestExtensionNextProtocolNegotiationClient(unittest.TestCase):
    def test_parse(self):
        extension_next_protocol_names_dict = collections.OrderedDict([
            ('extension_type', b'\x33\x74'),
            ('extension_length', b'\x00\x00'),
        ])
        extension_next_protocol_names_bytes = b''.join(extension_next_protocol_names_dict.values())
        extension_next_protocol_names_mac = TlsExtensionNextProtocolNegotiationClient.parse_exact_size(
            extension_next_protocol_names_bytes
        )
        self.assertEqual(extension_next_protocol_names_mac.compose(), extension_next_protocol_names_bytes)


class TestExtensionNextProtocolNegotiationServer(unittest.TestCase):
    def test_parse(self):
        extension_next_protocol_names_dict = collections.OrderedDict([
            ('extension_type', b'\x33\x74'),
            ('extension_length', b'\x00\x10'),
            ('protocol_name_h2_length', b'\x08'),
            ('protocol_name_h2', b'http/1.1'),
            ('protocol_name_h2c_length', b'\x06'),
            ('protocol_name_h2c', b'spdy/1'),
        ])
        extension_next_protocol_names_bytes = b''.join(extension_next_protocol_names_dict.values())
        extension_next_protocol_names = TlsExtensionNextProtocolNegotiationServer.parse_exact_size(
            extension_next_protocol_names_bytes
        )
        self.assertEqual(
            extension_next_protocol_names.protocol_names,
            TlsNextProtocolNameList([TlsNextProtocolName.HTTP_1_1, TlsNextProtocolName.SPDY_1])
        )
        self.assertEqual(extension_next_protocol_names.compose(), extension_next_protocol_names_bytes)


class TestExtensionApplicationLayerProtocolNegotiation(unittest.TestCase):
    def test_parse(self):
        extension_alpn_dict = collections.OrderedDict([
            ('extension_type', b'\x00\x10'),
            ('extension_length', b'\x00\x09'),
            ('protocol_name_list_length', b'\x00\x07'),
            ('protocol_name_h2_length', b'\x02'),
            ('protocol_name_h2', b'h2'),
            ('protocol_name_h2c_length', b'\x03'),
            ('protocol_name_h2c', b'h2c'),
        ])
        extension_alpn_bytes = b''.join(extension_alpn_dict.values())
        extension_alpn = TlsExtensionApplicationLayerProtocolNegotiation.parse_exact_size(
            extension_alpn_bytes
        )
        self.assertEqual(extension_alpn.protocol_names, TlsProtocolNameList([TlsProtocolName.H2, TlsProtocolName.H2C]))
        self.assertEqual(extension_alpn.compose(), extension_alpn_bytes)


class TestExtensionApplicationLayerProtocolSettings(unittest.TestCase):
    def test_parse(self):
        extension_alpn_dict = collections.OrderedDict([
            ('extension_type', b'\x44\x69'),
            ('extension_length', b'\x00\x09'),
            ('protocol_name_list_length', b'\x00\x07'),
            ('protocol_name_h2_length', b'\x02'),
            ('protocol_name_h2', b'h2'),
            ('protocol_name_h2c_length', b'\x03'),
            ('protocol_name_h2c', b'h2c'),
        ])
        extension_alpn_bytes = b''.join(extension_alpn_dict.values())
        extension_alpn = TlsExtensionApplicationLayerProtocolSettings.parse_exact_size(
            extension_alpn_bytes
        )
        self.assertEqual(extension_alpn.protocol_names, TlsProtocolNameList([TlsProtocolName.H2, TlsProtocolName.H2C]))
        self.assertEqual(extension_alpn.compose(), extension_alpn_bytes)


class TestExtensionUnusedData(unittest.TestCase):
    def test_error(self):
        extension_unused_data_dict = collections.OrderedDict([
            ('extension_type', b'\xff\x01'),
            ('extension_length', b'\x00\x01'),
            ('extension_data', b'\xff'),
        ])
        extension_unused_data_bytes = b''.join(extension_unused_data_dict.values())
        with self.assertRaises(InvalidValue) as context_manager:
            # pylint: disable=expression-not-assigned
            TestUnusedDataExtension.parse_exact_size(extension_unused_data_bytes)
        self.assertEqual(context_manager.exception.value, b'\xff')


class TestExtensionEncryptThenMAC(unittest.TestCase):
    def test_parse(self):
        extension_encrypt_then_mac_dict = collections.OrderedDict([
            ('extension_type', b'\x00\x16'),
            ('extension_length', b'\x00\x00'),
        ])
        extension_encrypt_then_mac_bytes = b''.join(extension_encrypt_then_mac_dict.values())
        extension_encrypt_then_mac = TlsExtensionEncryptThenMAC.parse_exact_size(extension_encrypt_then_mac_bytes)
        self.assertEqual(extension_encrypt_then_mac.compose(), extension_encrypt_then_mac_bytes)


class TestExtensionExtendedMasterSecret(unittest.TestCase):
    def test_parse(self):
        extension_extended_master_secret_dict = collections.OrderedDict([
            ('extension_type', b'\x00\x17'),
            ('extension_length', b'\x00\x00'),
        ])
        extension_extended_master_secret_bytes = b''.join(extension_extended_master_secret_dict.values())
        extended_master_secret = TlsExtensionExtendedMasterSecret.parse_exact_size(
            extension_extended_master_secret_bytes
        )
        self.assertEqual(extended_master_secret.compose(), extension_extended_master_secret_bytes)


class TestExtensionExtendedShortRecordHeader(unittest.TestCase):
    def test_parse(self):
        extension_short_record_header_dict = collections.OrderedDict([
            ('extension_type', b'\xff\x03'),
            ('extension_length', b'\x00\x00'),
        ])
        extension_short_record_header_bytes = b''.join(extension_short_record_header_dict.values())
        short_record_header = TlsExtensionShortRecordHeader.parse_exact_size(
            extension_short_record_header_bytes
        )
        self.assertEqual(short_record_header.compose(), extension_short_record_header_bytes)


class TestExtensionRecordSizeLimit(unittest.TestCase):
    def test_parse(self):
        extension_record_size_limit_dict = collections.OrderedDict([
            ('extension_type', b'\x00\x1c'),
            ('extension_length', b'\x00\x02'),
            ('record_size_limit', b'\x00\xff'),
        ])
        extension_record_size_limit_bytes = b''.join(extension_record_size_limit_dict.values())
        extension_record_size_limit = TlsExtensionRecordSizeLimit.parse_exact_size(extension_record_size_limit_bytes)
        self.assertEqual(extension_record_size_limit.record_size_limit, 0xff)
        self.assertEqual(extension_record_size_limit.compose(), extension_record_size_limit_bytes)


class TestExtensionPadding(unittest.TestCase):
    def test_error_non_zero_padding(self):
        extension_padding_dict = collections.OrderedDict([
            ('extension_type', b'\x00\x15'),
            ('extension_length', b'\x00\x04'),
            ('extension_data', b'\x00\x00\x00\x01'),
        ])
        extension_padding_bytes = b''.join(extension_padding_dict.values())

        with self.assertRaises(InvalidValue) as context_manager:
            TlsExtensionPadding.parse_exact_size(extension_padding_bytes)
        self.assertEqual(context_manager.exception.value, b'\x01')

    def test_parse(self):
        extension_padding_minimal_dict = collections.OrderedDict([
            ('extension_type', b'\x00\x15'),
            ('extension_length', b'\x00\x00'),
        ])
        extension_padding_minimal_bytes = b''.join(extension_padding_minimal_dict.values())

        extension_padding_minimal = TlsExtensionPadding.parse_exact_size(extension_padding_minimal_bytes)
        self.assertEqual(extension_padding_minimal.compose(), extension_padding_minimal_bytes)

        extension_padding_with_data_dict = collections.OrderedDict([
            ('extension_type', b'\x00\x15'),
            ('extension_length', b'\x00\x04'),
            ('extension_data', b'\x00\x00\x00\x00'),
        ])
        extension_padding_with_data_bytes = b''.join(extension_padding_with_data_dict.values())

        extension_padding_with_data = TlsExtensionPadding.parse_exact_size(extension_padding_with_data_bytes)
        self.assertEqual(extension_padding_with_data.compose(), extension_padding_with_data_bytes)


class ExtensionEncryptedClientHelloBase(unittest.TestCase):
    def test_parse(self):
        extension_encrypted_client_hello_dict = collections.OrderedDict([
            ('extension_type', b'\xfe\x0d'),
            ('extension_length', b'\x00\x01'),
            ('hello_type', b'\x00'),
        ])
        extension_encrypted_client_hello_bytes = b''.join(extension_encrypted_client_hello_dict.values())

        with self.assertRaises(InvalidType):
            # pylint: disable=expression-not-assigned
            TlsExtensionEncryptedClientHelloInner.parse_exact_size(extension_encrypted_client_hello_bytes)


class ExtensionEncryptedClientHelloInner(unittest.TestCase):
    def test_parse(self):
        extension_encrypted_client_hello_dict = collections.OrderedDict([
            ('extension_type', b'\xfe\x0d'),
            ('extension_length', b'\x00\x01'),
            ('hello_type', b'\x01'),
        ])
        extension_encrypted_client_hello_bytes = b''.join(extension_encrypted_client_hello_dict.values())

        extension_encrypted_client_hello = TlsExtensionEncryptedClientHelloInner.parse_exact_size(
            extension_encrypted_client_hello_bytes
        )
        self.assertEqual(
            extension_encrypted_client_hello.compose(),
            extension_encrypted_client_hello_bytes
        )


class ExtensionEncryptedClientHelloOuter(unittest.TestCase):
    def test_parse(self):
        extension_encrypted_client_hello_dict = collections.OrderedDict([
            ('extension_type', b'\xfe\x0d'),
            ('extension_length', b'\x00\x11'),
            ('hello_type', b'\x00'),
            ('hello_data', b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'),
        ])
        extension_encrypted_client_hello_bytes = b''.join(extension_encrypted_client_hello_dict.values())

        extension_encrypted_client_hello = TlsExtensionEncryptedClientHelloOuter.parse_exact_size(
            extension_encrypted_client_hello_bytes
        )
        self.assertEqual(
            extension_encrypted_client_hello.compose(),
            extension_encrypted_client_hello_bytes
        )


class TestExtensionPostHandshakeAuthentication(unittest.TestCase):
    def test_parse(self):
        extension_post_handshake_authentication_dict = collections.OrderedDict([
            ('extension_type', b'\x00\x31'),
            ('extension_length', b'\x00\x00'),
        ])
        extension_post_handshake_authentication_bytes = b''.join(extension_post_handshake_authentication_dict.values())

        extension_post_handshake_authentication = TlsExtensionPostHandshakeAuthentication.parse_exact_size(
            extension_post_handshake_authentication_bytes
        )
        self.assertEqual(
            extension_post_handshake_authentication.compose(),
            extension_post_handshake_authentication_bytes
        )



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import unittest

from cryptoparser.common.utils import bytes_from_hex_string, bytes_to_hex_string


class TestBytesToHexString(unittest.TestCase):
    def test_separator(self):
        self.assertEqual(bytes_to_hex_string(b''), '')
        self.assertEqual(bytes_to_hex_string(b'\xde\xad\xbe\xef'), 'DEADBEEF')
        self.assertEqual(bytes_to_hex_string(b'\xde\xad\xbe\xef', separator=':'), 'DE:AD:BE:EF')

    def test_lowercase(self):
        self.assertEqual(bytes_to_hex_string(b''), '')
        self.assertEqual(bytes_to_hex_string(b'\xde\xad\xbe\xef'), 'DEADBEEF')
        self.assertEqual(bytes_to_hex_string(b'\xde\xad\xbe\xef', lowercase=True), 'deadbeef')


class TestBytesFromHexString(unittest.TestCase):
    def test_error_odd_length_string(self):
        with self.assertRaises(ValueError) as context_manager:
            bytes_from_hex_string('0d:d')
        self.assertEqual(type(context_manager.exception), ValueError)

    def test_error_non_hex_string(self):
        with self.assertRaises(ValueError) as context_manager:
            bytes_from_hex_string('no:th:ex')
        self.assertEqual(type(context_manager.exception), ValueError)

    def test_separator(self):
        self.assertEqual(bytes_from_hex_string(''), b'')
        self.assertEqual(bytes_from_hex_string('DEADBEEF'), b'\xde\xad\xbe\xef')
        self.assertEqual(bytes_from_hex_string('DE:AD:BE:EF', separator=':'), b'\xde\xad\xbe\xef')



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import unittest

from cryptoparser.httpx.version import HttpVersion


class TestHttpVersion(unittest.TestCase):
    def test_markdown(self):
        self.assertEqual(HttpVersion.HTTP1_0.value.as_json(), '"http1_0"')
        self.assertEqual(HttpVersion.HTTP1_0.value.as_markdown(), 'HTTP/1.0')

        self.assertEqual(HttpVersion.HTTP1_1.value.as_json(), '"http1_1"')
        self.assertEqual(HttpVersion.HTTP1_1.value.as_markdown(), 'HTTP/1.1')



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import collections
import unittest

from cryptodatahub.ike.algorithm import (
    Ikev1PayloadType, Ikev1AuthenticationMethod, Ikev1TransformId, Ikev1ProtocolId, Ikev1Doi
)

from cryptoparser.ike.ikev1 import (
    Ikev1AttributeAuthenticationMethod, Ikev1PayloadTransform, Ikev1PayloadProposal,
    Ikev1PayloadSecurityAssociation, Ikev1Situation
)


class TestIkev1PayloadProposal(unittest.TestCase):  # pylint: disable=too-many-instance-attributes
    def setUp(self):
        self.simple_transform = Ikev1PayloadTransform(
            transform_id=Ikev1TransformId.KEY_IKE,
            attributes=[]
        )

        self.auth_attribute = Ikev1AttributeAuthenticationMethod(value=Ikev1AuthenticationMethod.PRE_SHARED_KEY)
        self.transform_with_attr = Ikev1PayloadTransform(
            transform_id=Ikev1TransformId.KEY_IKE,
            attributes=[self.auth_attribute]
        )

        self.test_dict_single_transform = collections.OrderedDict([
            ('next_payload', b'\x00'),  # NONE = 0x00
            ('reserved', b'\x00'),
            ('payload_length', b'\x00\x10'),  # 16 bytes total
            ('proposal_number', b'\x01'),
            ('protocol_id', b'\x01'),  # ISAKMP = 0x01
            ('spi_size', b'\x00'),
            ('transform_count', b'\x01'),
            ('transform_data', b'\x00\x00\x00\x08\x01\x01\x00\x00'),  # 8-byte transform
        ])
        self.test_bytes_single_transform = b''.join(self.test_dict_single_transform.values())

        self.spi_data = b'\x12\x34\x56\x78'
        self.test_dict_with_spi = collections.OrderedDict([
            ('next_payload', b'\x00'),  # NONE = 0x00
            ('reserved', b'\x00'),
            ('payload_length', b'\x00\x14'),  # 20 bytes total
            ('proposal_number', b'\x01'),
            ('protocol_id', b'\x03'),  # IPSEC_ESP = 0x03
            ('spi_size', b'\x04'),  # 4-byte SPI
            ('transform_count', b'\x01'),
            ('spi', self.spi_data),
            ('transform_data', b'\x00\x00\x00\x08\x01\x01\x00\x00'),  # 8-byte transform
        ])
        self.test_bytes_with_spi = b''.join(self.test_dict_with_spi.values())

        self.test_dict_multi_transforms = collections.OrderedDict([
            ('next_payload', b'\x00'),  # NONE = 0x00
            ('reserved', b'\x00'),
            ('payload_length', b'\x00\x18'),  # 24 bytes total
            ('proposal_number', b'\x01'),
            ('protocol_id', b'\x01'),  # ISAKMP = 0x01
            ('spi_size', b'\x00'),
            ('transform_count', b'\x02'),
            ('transform1_data', b'\x03\x00\x00\x08\x01\x01\x00\x00'),  # next=TRANSFORM
            ('transform2_data', b'\x00\x00\x00\x08\x02\x01\x00\x00'),  # next=NONE
        ])
        self.test_bytes_multi_transforms = b''.join(self.test_dict_multi_transforms.values())

        self.proposal_single = Ikev1PayloadProposal(
            protocol_id=Ikev1ProtocolId.ISAKMP,
            transforms=[self.simple_transform],
            spi=b''
        )
        self.proposal_single.proposal_number = 1
        self.proposal_single.next_payload = Ikev1PayloadType.NONE

        self.proposal_with_spi = Ikev1PayloadProposal(
            protocol_id=Ikev1ProtocolId.IPSEC_ESP,
            transforms=[self.simple_transform],
            spi=self.spi_data
        )
        self.proposal_with_spi.proposal_number = 1
        self.proposal_with_spi.next_payload = Ikev1PayloadType.NONE

        self.second_transform = Ikev1PayloadTransform(
            transform_id=Ikev1TransformId.KEY_IKE,
            attributes=[]
        )

        self.proposal_multi = Ikev1PayloadProposal(
            protocol_id=Ikev1ProtocolId.ISAKMP,
            transforms=[self.simple_transform, self.second_transform],
            spi=b''
        )
        self.proposal_multi.proposal_number = 1
        self.proposal_multi.next_payload = Ikev1PayloadType.NONE

    def test_get_payload_type(self):
        self.assertEqual(Ikev1PayloadProposal.get_payload_type(), Ikev1PayloadType.PROPOSAL)

    def test_constructor_with_single_transform(self):
        proposal = Ikev1PayloadProposal(
            protocol_id=Ikev1ProtocolId.ISAKMP,
            transforms=[self.simple_transform],
            spi=b''
        )
        self.assertEqual(proposal.protocol_id, Ikev1ProtocolId.ISAKMP)
        self.assertEqual(len(proposal.transforms), 1)
        self.assertEqual(proposal.transforms[0], self.simple_transform)
        self.assertEqual(proposal.spi, b'')

    def test_constructor_with_spi(self):
        proposal = Ikev1PayloadProposal(
            protocol_id=Ikev1ProtocolId.IPSEC_ESP,
            transforms=[self.simple_transform],
            spi=self.spi_data
        )
        self.assertEqual(proposal.protocol_id, Ikev1ProtocolId.IPSEC_ESP)
        self.assertEqual(proposal.spi, self.spi_data)

    def test_constructor_with_multiple_transforms(self):
        proposal = Ikev1PayloadProposal(
            protocol_id=Ikev1ProtocolId.ISAKMP,
            transforms=[self.simple_transform, self.second_transform],
            spi=b''
        )
        self.assertEqual(len(proposal.transforms), 2)
        self.assertEqual(proposal.transforms[0], self.simple_transform)
        self.assertEqual(proposal.transforms[1], self.second_transform)

    def test_parse_single_transform(self):
        parsed: Ikev1PayloadProposal = Ikev1PayloadProposal.parse_exact_size(self.test_bytes_single_transform)

        self.assertEqual(parsed.next_payload, Ikev1PayloadType.NONE)
        self.assertEqual(parsed.proposal_number, 1)
        self.assertEqual(parsed.protocol_id, Ikev1ProtocolId.ISAKMP)
        self.assertEqual(parsed.spi, b'')
        self.assertEqual(len(parsed.transforms), 1)
        self.assertEqual(parsed.transforms[0].transform_id, Ikev1TransformId.KEY_IKE)

    def test_parse_with_spi(self):
        parsed: Ikev1PayloadProposal = Ikev1PayloadProposal.parse_exact_size(self.test_bytes_with_spi)

        self.assertEqual(parsed.protocol_id, Ikev1ProtocolId.IPSEC_ESP)
        self.assertEqual(parsed.spi, self.spi_data)
        self.assertEqual(len(parsed.transforms), 1)

    def test_parse_multiple_transforms(self):
        parsed: Ikev1PayloadProposal = Ikev1PayloadProposal.parse_exact_size(self.test_bytes_multi_transforms)

        self.assertEqual(parsed.proposal_number, 1)
        self.assertEqual(parsed.protocol_id, Ikev1ProtocolId.ISAKMP)
        self.assertEqual(len(parsed.transforms), 2)
        self.assertEqual(parsed.transforms[0].transform_id, Ikev1TransformId.KEY_IKE)
        self.assertEqual(parsed.transforms[1].transform_id, Ikev1TransformId.KEY_IKE)

    def test_compose_single_transform(self):
        composed_bytes = self.proposal_single.compose()
        self.assertEqual(composed_bytes, self.test_bytes_single_transform)

    def test_compose_with_spi(self):
        composed_bytes = self.proposal_with_spi.compose()
        self.assertEqual(composed_bytes, self.test_bytes_with_spi)

    def test_compose_multiple_transforms(self):
        composed_bytes = self.proposal_multi.compose()
        self.assertEqual(composed_bytes, self.test_bytes_multi_transforms)

    def test_round_trip_single_transform(self):
        composed_bytes = self.proposal_single.compose()
        parsed: Ikev1PayloadProposal = Ikev1PayloadProposal.parse_exact_size(composed_bytes)

        self.assertEqual(parsed.protocol_id, self.proposal_single.protocol_id)
        self.assertEqual(parsed.spi, self.proposal_single.spi)
        self.assertEqual(len(parsed.transforms), len(self.proposal_single.transforms))
        self.assertEqual(parsed.proposal_number, self.proposal_single.proposal_number)
        self.assertEqual(parsed.next_payload, self.proposal_single.next_payload)

    def test_round_trip_with_spi(self):
        composed_bytes = self.proposal_with_spi.compose()
        parsed: Ikev1PayloadProposal = Ikev1PayloadProposal.parse_exact_size(composed_bytes)

        self.assertEqual(parsed.protocol_id, self.proposal_with_spi.protocol_id)
        self.assertEqual(parsed.spi, self.proposal_with_spi.spi)
        self.assertEqual(len(parsed.transforms), len(self.proposal_with_spi.transforms))

    def test_round_trip_multiple_transforms(self):
        composed_bytes = self.proposal_multi.compose()
        parsed: Ikev1PayloadProposal = Ikev1PayloadProposal.parse_exact_size(composed_bytes)

        self.assertEqual(parsed.protocol_id, self.proposal_multi.protocol_id)
        self.assertEqual(len(parsed.transforms), len(self.proposal_multi.transforms))
        self.assertEqual(parsed.transforms[0].next_payload, Ikev1PayloadType.TRANSFORM)
        self.assertEqual(parsed.transforms[1].next_payload, Ikev1PayloadType.NONE)

    def test_transform_numbering_in_compose(self):
        composed_bytes = self.proposal_multi.compose()

        self.assertEqual(composed_bytes[12], 1)  # First transform number at byte 12
        self.assertEqual(composed_bytes[20], 2)  # Second transform number at byte 20

    def test_spi_size_calculation(self):
        different_spi = b'\xaa\xbb\xcc\xdd\xee\xff'  # 6 bytes
        proposal = Ikev1PayloadProposal(
            protocol_id=Ikev1ProtocolId.IPSEC_AH,
            transforms=[self.simple_transform],
            spi=different_spi
        )
        proposal.proposal_number = 1
        proposal.next_payload = Ikev1PayloadType.NONE

        composed_bytes = proposal.compose()
        self.assertEqual(composed_bytes[6], 6)  # SPI size field at byte 6


class TestIkev1PayloadSecurityAssociation(unittest.TestCase):  # pylint: disable=too-many-instance-attributes
    def setUp(self):
        # Simple transform for proposals
        self.simple_transform = Ikev1PayloadTransform(
            transform_id=Ikev1TransformId.KEY_IKE,
            attributes=[]
        )

        # Simple proposal for SA
        self.simple_proposal = Ikev1PayloadProposal(
            protocol_id=Ikev1ProtocolId.ISAKMP,
            transforms=[self.simple_transform],
            spi=b''
        )

        # Second proposal for multi-proposal tests
        self.second_proposal = Ikev1PayloadProposal(
            protocol_id=Ikev1ProtocolId.IPSEC_ESP,
            transforms=[self.simple_transform],
            spi=b'\x12\x34\x56\x78'
        )

        # Test data for SA with single proposal
        self.test_dict_single_proposal = collections.OrderedDict([
            ('next_payload', b'\x00'),  # NONE = 0x00
            ('reserved', b'\x00'),
            ('payload_length', b'\x00\x1c'),  # 28 bytes total
            ('doi', b'\x00\x00\x00\x01'),  # IPSEC = 0x00000001
            ('situation', b'\x00\x00\x00\x01'),  # SIT_IDENTITY_ONLY = 0x00000001
            # Proposal: next=NONE, length=16, prop_num=1, protocol=ISAKMP, spi_size=0, transform_count=1
            ('proposal_data', b'\x00\x00\x00\x10\x01\x01\x00\x01'),
            # Transform: next=NONE, length=8, transform_num=1, transform_id=KEY_IKE
            ('transform_data', b'\x00\x00\x00\x08\x01\x01\x00\x00'),
        ])
        self.test_bytes_single_proposal = b''.join(self.test_dict_single_proposal.values())

        # Test data for SA with multiple proposals
        self.test_dict_multi_proposals = collections.OrderedDict([
            ('next_payload', b'\x00'),  # NONE = 0x00
            ('reserved', b'\x00'),
            ('payload_length', b'\x00\x30'),  # 48 bytes total
            ('doi', b'\x00\x00\x00\x01'),  # IPSEC = 0x00000001
            ('situation', b'\x00\x00\x00\x01'),  # SIT_IDENTITY_ONLY = 0x00000001
            # First proposal: next=PROPOSAL, length=16, prop_num=1
            ('proposal1_data', b'\x02\x00\x00\x10\x01\x01\x00\x01'),
            ('transform1_data', b'\x00\x00\x00\x08\x01\x01\x00\x00'),
            # Second proposal: next=NONE, length=20, prop_num=2, protocol=ESP, 4-byte SPI
            ('proposal2_data', b'\x00\x00\x00\x14\x02\x03\x04\x01'),
            ('spi_data', b'\x12\x34\x56\x78'),
            ('transform2_data', b'\x00\x00\x00\x08\x01\x01\x00\x00'),
        ])
        self.test_bytes_multi_proposals = b''.join(self.test_dict_multi_proposals.values())

        # Test data for different situation flags
        self.test_dict_secrecy_integrity = collections.OrderedDict([
            ('next_payload', b'\x00'),  # NONE = 0x00
            ('reserved', b'\x00'),
            ('payload_length', b'\x00\x1c'),  # 28 bytes total
            ('doi', b'\x00\x00\x00\x01'),  # IPSEC = 0x00000001
            ('situation', b'\x00\x00\x00\x06'),  # SIT_SECRECY | SIT_INTEGRITY = 0x02 | 0x04 = 0x06
            ('proposal_data', b'\x00\x00\x00\x10\x01\x01\x00\x01'),
            ('transform_data', b'\x00\x00\x00\x08\x01\x01\x00\x00'),
        ])
        self.test_bytes_secrecy_integrity = b''.join(self.test_dict_secrecy_integrity.values())

        # SA objects for compose tests
        self.sa_single = Ikev1PayloadSecurityAssociation(
            doi=Ikev1Doi.IPSEC,
            situation={Ikev1Situation.SIT_IDENTITY_ONLY},
            proposals=[self.simple_proposal]
        )
        self.sa_single.next_payload = Ikev1PayloadType.NONE

        self.sa_multi = Ikev1PayloadSecurityAssociation(
            doi=Ikev1Doi.IPSEC,
            situation={Ikev1Situation.SIT_IDENTITY_ONLY},
            proposals=[self.simple_proposal, self.second_proposal]
        )
        self.sa_multi.next_payload = Ikev1PayloadType.NONE

        self.sa_flags = Ikev1PayloadSecurityAssociation(
            doi=Ikev1Doi.IPSEC,
            situation={Ikev1Situation.SIT_SECRECY, Ikev1Situation.SIT_INTEGRITY},
            proposals=[self.simple_proposal]
        )
        self.sa_flags.next_payload = Ikev1PayloadType.NONE

    def test_get_payload_type(self):
        self.assertEqual(Ikev1PayloadSecurityAssociation.get_payload_type(), Ikev1PayloadType.SECURITY_ASSOCIATION)

    def test_constructor_with_single_proposal(self):
        sa = Ikev1PayloadSecurityAssociation(
            doi=Ikev1Doi.IPSEC,
            situation={Ikev1Situation.SIT_IDENTITY_ONLY},
            proposals=[self.simple_proposal]
        )
        self.assertEqual(sa.doi, Ikev1Doi.IPSEC)
        self.assertEqual(sa.situation, {Ikev1Situation.SIT_IDENTITY_ONLY})
        self.assertEqual(len(sa.proposals), 1)
        self.assertEqual(sa.proposals[0], self.simple_proposal)

    def test_constructor_with_multiple_proposals(self):
        sa = Ikev1PayloadSecurityAssociation(
            doi=Ikev1Doi.IPSEC,
            situation={Ikev1Situation.SIT_IDENTITY_ONLY},
            proposals=[self.simple_proposal, self.second_proposal]
        )
        self.assertEqual(len(sa.proposals), 2)
        self.assertEqual(sa.proposals[0], self.simple_proposal)
        self.assertEqual(sa.proposals[1], self.second_proposal)

    def test_constructor_with_situation_flags(self):
        sa = Ikev1PayloadSecurityAssociation(
            doi=Ikev1Doi.IPSEC,
            situation={Ikev1Situation.SIT_SECRECY, Ikev1Situation.SIT_INTEGRITY},
            proposals=[self.simple_proposal]
        )
        self.assertEqual(sa.situation, {Ikev1Situation.SIT_SECRECY, Ikev1Situation.SIT_INTEGRITY})

    def test_doi_storage(self):
        sa = Ikev1PayloadSecurityAssociation(
            doi=Ikev1Doi.GDOI,
            situation={Ikev1Situation.SIT_IDENTITY_ONLY},
            proposals=[self.simple_proposal]
        )
        self.assertEqual(sa.doi, Ikev1Doi.GDOI)

    def test_parse_single_proposal(self):
        parsed: Ikev1PayloadSecurityAssociation = Ikev1PayloadSecurityAssociation.parse_exact_size(
            self.test_bytes_single_proposal
        )

        self.assertEqual(parsed.next_payload, Ikev1PayloadType.NONE)
        self.assertEqual(parsed.doi, Ikev1Doi.IPSEC)
        self.assertEqual(parsed.situation, {Ikev1Situation.SIT_IDENTITY_ONLY})
        self.assertEqual(len(parsed.proposals), 1)
        self.assertEqual(parsed.proposals[0].protocol_id, Ikev1ProtocolId.ISAKMP)

    def test_parse_multiple_proposals(self):
        parsed: Ikev1PayloadSecurityAssociation = Ikev1PayloadSecurityAssociation.parse_exact_size(
            self.test_bytes_multi_proposals
        )

        self.assertEqual(parsed.doi, Ikev1Doi.IPSEC)
        self.assertEqual(len(parsed.proposals), 2)
        self.assertEqual(parsed.proposals[0].protocol_id, Ikev1ProtocolId.ISAKMP)
        self.assertEqual(parsed.proposals[1].protocol_id, Ikev1ProtocolId.IPSEC_ESP)

    def test_parse_situation_flags(self):
        parsed: Ikev1PayloadSecurityAssociation = Ikev1PayloadSecurityAssociation.parse_exact_size(
            self.test_bytes_secrecy_integrity
        )

        self.assertEqual(parsed.situation, {Ikev1Situation.SIT_SECRECY, Ikev1Situation.SIT_INTEGRITY})

    def test_compose_single_proposal(self):
        composed_bytes = self.sa_single.compose()
        self.assertEqual(composed_bytes, self.test_bytes_single_proposal)

    def test_compose_multiple_proposals(self):
        composed_bytes = self.sa_multi.compose()
        self.assertEqual(composed_bytes, self.test_bytes_multi_proposals)

    def test_compose_situation_flags(self):
        composed_bytes = self.sa_flags.compose()
        self.assertEqual(composed_bytes, self.test_bytes_secrecy_integrity)

    def test_round_trip_single_proposal(self):
        composed_bytes = self.sa_single.compose()
        parsed: Ikev1PayloadSecurityAssociation = Ikev1PayloadSecurityAssociation.parse_exact_size(composed_bytes)

        self.assertEqual(parsed.doi, self.sa_single.doi)
        self.assertEqual(parsed.situation, self.sa_single.situation)
        self.assertEqual(len(parsed.proposals), len(self.sa_single.proposals))
        self.assertEqual(parsed.next_payload, self.sa_single.next_payload)

    def test_round_trip_multiple_proposals(self):
        composed_bytes = self.sa_multi.compose()
        parsed: Ikev1PayloadSecurityAssociation = Ikev1PayloadSecurityAssociation.parse_exact_size(composed_bytes)

        self.assertEqual(parsed.doi, self.sa_multi.doi)
        self.assertEqual(len(parsed.proposals), len(self.sa_multi.proposals))
        # Verify proposal chaining (first proposal points to next, last to NONE)
        self.assertEqual(parsed.proposals[0].next_payload, Ikev1PayloadType.PROPOSAL)
        self.assertEqual(parsed.proposals[1].next_payload, Ikev1PayloadType.NONE)

    def test_round_trip_situation_flags(self):
        composed_bytes = self.sa_flags.compose()
        parsed: Ikev1PayloadSecurityAssociation = Ikev1PayloadSecurityAssociation.parse_exact_size(composed_bytes)

        self.assertEqual(parsed.situation, self.sa_flags.situation)

    def test_proposal_numbering_in_compose(self):
        composed_bytes = self.sa_multi.compose()

        # Verify proposal numbers are written correctly (1-based indexing)
        self.assertEqual(composed_bytes[16], 1)  # First proposal number at byte 16
        self.assertEqual(composed_bytes[32], 2)  # Second proposal number at byte 32


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import json
import unittest


from cryptodatahub.common.exception import InvalidValue

from cryptoparser.common.exception import NotEnoughData, TooMuchData
from cryptoparser.common.base import (
    Opaque,
    OpaqueEnumParsable,
    OpaqueParam,
    ProtocolVersionMajorMinorBase,
    Serializable,
    SerializableTextEncoder,
    Vector,
    VectorEnumCodeString,
    VectorParamEnumCodeString,
    VectorParamNumeric,
    VectorParamParsable,
    VectorParamString,
    VectorParsable,
    VectorParsableDerived,
    VectorString,
)
from cryptoparser.common.parse import ComposerBinary

from .classes import (
    AlwaysTestStringComposer,
    ConditionalParsable,
    EnumStringValue,
    FourByteEnumComposerTest,
    FourByteEnumParsableTest,
    ListParsableTest,
    NByteEnumTest,
    NumericRangeParsableTest,
    OneByteEnumComposerTest,
    OneByteEnumParsableTest,
    OneByteOddParsable,
    OneByteParsable,
    OpaqueEnum,
    OpaqueEnumFactory,
    SerializableAttributeOrder,
    SerializableEmptyValues,
    SerializableEnums,
    SerializableHidden,
    SerializableHumanFriendly,
    SerializableHumanReadable,
    SerializableIterables,
    SerializableRecursive,
    SerializableSimpleTypes,
    SerializableSingle,
    SerializableUnhandled,
    SerializableUpperCaseEncoder,
    StringEnum,
    StringEnumA,
    StringEnumAA,
    StringEnumAAA,
    TestObject,
    ThreeByteEnumComposerTest,
    ThreeByteEnumParsableTest,
    TwoByteEnumComposerTest,
    TwoByteEnumParsableTest,
    TwoByteEvenParsable,
    TwoByteParsable,
    VariantParsableTest,
    VariantParsableExactTest,
)


class TestProtocolVersionMajorMinorBase(unittest.TestCase):
    def setUp(self):
        self.protocol_version = ProtocolVersionMajorMinorBase(1, 2)
        self.protocol_version_bytes = b'\x01\x02'

    def test_error(self):
        with self.assertRaises(NotEnoughData) as context_manager:
            ProtocolVersionMajorMinorBase.parse_exact_size(b'\x00')
        self.assertEqual(context_manager.exception.bytes_needed, 1)

    def test_parse(self):
        self.assertEqual(
            ProtocolVersionMajorMinorBase.parse_exact_size(self.protocol_version_bytes),
            self.protocol_version
        )

    def test_compose(self):
        self.assertEqual(self.protocol_version.compose(), self.protocol_version_bytes)

    def test_identifier(self):
        self.assertEqual(self.protocol_version.identifier, '1_2')

    def test_markdown(self):
        self.assertEqual(self.protocol_version.as_markdown(), '1.2')

    def test_str(self):
        self.assertEqual(str(self.protocol_version), '1.2')


class VectorNumericTestErrors(Vector):
    @classmethod
    def get_param(cls):
        return VectorParamNumeric(item_size=2, min_byte_num=4, max_byte_num=6)


class VectorNumericTest(Vector):
    @classmethod
    def get_param(cls):
        return VectorParamNumeric(item_size=2, min_byte_num=0, max_byte_num=0xff)


class VectorStringTest(VectorString):
    @classmethod
    def get_param(cls):
        return VectorParamString(
            min_byte_num=0,
            max_byte_num=16,
            separator=';',
            item_class=StringEnum,
            fallback_class=str
        )


class VectorOneByteParsableTest(VectorParsable):
    @classmethod
    def get_param(cls):
        return VectorParamParsable(item_class=OneByteParsable, min_byte_num=0, max_byte_num=0xff, fallback_class=None)


class VectorTwoByteParsableTest(VectorParsable):
    @classmethod
    def get_param(cls):
        return VectorParamParsable(item_class=TwoByteParsable, min_byte_num=0, max_byte_num=0xffff, fallback_class=None)


class VectorConsditionalParsableTest(VectorParsableDerived):
    @classmethod
    def get_param(cls):
        return VectorParamParsable(
            item_class=ConditionalParsable,
            min_byte_num=0,
            max_byte_num=0xff,
            fallback_class=None
        )


class VectorFallbackParsableTest(VectorParsableDerived):
    @classmethod
    def get_param(cls):
        return VectorParamParsable(
            item_class=OneByteOddParsable,
            min_byte_num=0,
            max_byte_num=0xff,
            fallback_class=TwoByteEvenParsable
        )


class OpaqueTest(Opaque):
    @classmethod
    def get_param(cls):
        return OpaqueParam(min_byte_num=3, max_byte_num=3)


class TestVectorNumeric(unittest.TestCase):
    def test_error(self):
        with self.assertRaises(NotEnoughData) as context_manager:
            VectorNumericTestErrors(items=[1, ])
        self.assertEqual(context_manager.exception.bytes_needed, VectorNumericTestErrors.get_param().min_byte_num)

        with self.assertRaises(TooMuchData) as context_manager:
            VectorNumericTestErrors(items=[1, 2, 3, 4, ])
        self.assertEqual(context_manager.exception.bytes_needed, VectorNumericTestErrors.get_param().max_byte_num)

        vector = VectorNumericTestErrors(items=[1, 2, ])
        with self.assertRaises(NotEnoughData) as context_manager:
            del vector[0]
        self.assertEqual(context_manager.exception.bytes_needed, VectorNumericTestErrors.get_param().min_byte_num)

        vector = VectorNumericTestErrors(items=[1, 2, 3])
        with self.assertRaises(TooMuchData) as context_manager:
            vector.append(0xff)
        self.assertEqual(context_manager.exception.bytes_needed, VectorNumericTestErrors.get_param().max_byte_num)

    def test_parse(self):
        self.assertEqual(len(VectorNumericTest.parse_exact_size(b'\x00')), 0)

        self.assertEqual(
            [1, 2, ],
            list(VectorNumericTest.parse_exact_size(b'\x04\x00\x01\x00\x02'))
        )

    def test_compose(self):
        self.assertEqual(
            b'\x00',
            VectorNumericTest([]).compose(),
        )

        self.assertEqual(
            b'\x04\x00\x01\x00\x02',
            VectorNumericTest([1, 2, ]).compose(),
        )

    def test_container(self):
        vector = VectorNumericTest(items=[])

        vector.append(1)
        self.assertEqual(vector[0], 1)
        self.assertEqual(len(vector), 1)
        self.assertEqual(str(vector), '[1]')

        vector.insert(0, 0)
        self.assertEqual(vector[0], 0)
        self.assertEqual(len(vector), 2)
        self.assertEqual(str(vector), '[0, 1]')

        del vector[0]
        self.assertEqual(vector[0], 1)
        self.assertEqual(len(vector), 1)
        self.assertEqual(str(vector), '[1]')

        vector[0] = 0
        self.assertEqual(vector[0], 0)
        self.assertEqual(len(vector), 1)
        self.assertEqual(str(vector), '[0]')


class TestVectorString(unittest.TestCase):
    def test_error(self):
        pass

    def test_parse(self):
        self.assertEqual(len(VectorStringTest.parse_exact_size(b'\x00')), 0)

        self.assertEqual(
            [StringEnum.ONE, StringEnum.TWO, StringEnum.THREE, ],
            list(VectorStringTest.parse_exact_size(b'\x0fone;two;three'))
        )
        self.assertEqual(
            [StringEnum.ONE, StringEnum.TWO, StringEnum.THREE, 'four', ],
            list(VectorStringTest.parse_exact_size(b'\x14one;two;three;four'))
        )

    def test_compose(self):
        self.assertEqual(
            b'\x00',
            VectorStringTest([]).compose(),
        )

        self.assertEqual(
            b'\x07one;two',
            VectorStringTest([StringEnum.ONE, StringEnum.TWO, ]).compose(),
        )

    def test_json(self):
        self.assertEqual('[]', VectorStringTest([]).as_json())

        self.assertEqual(
            VectorStringTest([StringEnum.ONE, StringEnum.TWO, ]).as_json(),
            '[{"ONE": {"code": "one"}}, {"TWO": {"code": "two"}}]',
        )

    def test_markdown(self):
        self.assertEqual('-', VectorStringTest([]).as_markdown())

        self.assertEqual(
            VectorStringTest([StringEnum.ONE, StringEnum.TWO, ]).as_markdown(),
            '\n'.join([
                '1. ONE',
                '2. TWO',
                '',
            ])
        )


class StringEnumFactory(OpaqueEnumParsable):
    @classmethod
    def get_enum_class(cls):
        return StringEnum

    @classmethod
    def get_param(cls):
        return OpaqueParam(
            min_byte_num=1, max_byte_num=2 ** 8 - 1
        )


class VectorEnumCodeStringTest(VectorEnumCodeString):
    @classmethod
    def get_param(cls):
        return VectorParamEnumCodeString(
            item_class=StringEnumFactory,
            min_byte_num=0, max_byte_num=2 ** 16 - 1
        )


class TestVectorEnumCodeString(unittest.TestCase):
    def test_error(self):
        pass

    def test_parse(self):
        self.assertEqual(
            [StringEnum.ONE, StringEnum.TWO, StringEnum.THREE, ],
            list(VectorEnumCodeStringTest.parse_exact_size(b'\x00\x0e\x03one\x03two\x05three'))
        )

    def test_compose(self):
        self.assertEqual(
            b'\x00\x00',
            VectorEnumCodeStringTest([]).compose(),
        )

        self.assertEqual(
            b'\x00\x08\x03one\x03two',
            VectorEnumCodeStringTest([StringEnum.ONE, StringEnum.TWO, ]).compose(),
        )

    def test_json(self):
        self.assertEqual('[]', VectorEnumCodeStringTest([]).as_json())

        self.assertEqual(
            VectorEnumCodeStringTest([StringEnum.ONE, StringEnum.TWO, ]).as_json(),
            '[{"ONE": {"code": "one"}}, {"TWO": {"code": "two"}}]',
        )

    def test_markdown(self):
        self.assertEqual('-', VectorEnumCodeStringTest([]).as_markdown())

        self.assertEqual(
            VectorEnumCodeStringTest([StringEnum.ONE, StringEnum.TWO, ]).as_markdown(),
            '\n'.join([
                '1. ONE',
                '2. TWO',
                '',
            ])
        )


class TestVectorParsable(unittest.TestCase):
    def test_error(self):
        with self.assertRaises(NotEnoughData) as context_manager:
            VectorOneByteParsableTest.parse_exact_size(b'')
        self.assertEqual(context_manager.exception.bytes_needed, 1)

        with self.assertRaises(NotEnoughData) as context_manager:
            VectorTwoByteParsableTest.parse_exact_size(b'\x00')
        self.assertEqual(context_manager.exception.bytes_needed, 1)

        with self.assertRaises(NotEnoughData) as context_manager:
            VectorOneByteParsableTest.parse_exact_size(b'\x03')
        self.assertEqual(context_manager.exception.bytes_needed, 3)

        with self.assertRaises(NotEnoughData) as context_manager:
            VectorTwoByteParsableTest.parse_exact_size(b'\x00\x03')
        self.assertEqual(context_manager.exception.bytes_needed, 3)

    def test_parse(self):
        self.assertEqual(len(VectorOneByteParsableTest.parse_exact_size(b'\x00')), 0)
        self.assertEqual(len(VectorTwoByteParsableTest.parse_exact_size(b'\x00\x00')), 0)

        self.assertEqual(
            [0, 1, 0, 2, ],
            list(map(int, VectorOneByteParsableTest.parse_exact_size(b'\x04\x00\x01\x00\x02')))
        )
        self.assertEqual(
            [1, 2, ],
            list(map(int, VectorTwoByteParsableTest.parse_exact_size(b'\x00\x04\x00\x01\x00\x02')))
        )

        self.assertEqual(
            [0x01, 0x0200],
            list(map(int, VectorFallbackParsableTest.parse_exact_size(b'\x03\x01\x02\x00')))
        )

    def test_compose(self):
        self.assertEqual(b'\x00', VectorOneByteParsableTest([]).compose())
        self.assertEqual(b'\x00\x00', VectorTwoByteParsableTest([]).compose())

        self.assertEqual(
            b'\x04\x00\x01\x00\x02',
            VectorOneByteParsableTest([
                OneByteParsable(0),
                OneByteParsable(1),
                OneByteParsable(0),
                OneByteParsable(2)
            ]).compose(),
        )
        self.assertEqual(
            b'\x00\x04\x00\x01\x00\x02',
            VectorTwoByteParsableTest([
                TwoByteParsable(1),
                TwoByteParsable(2)
            ]).compose(),
        )

    def test_container(self):
        vector = VectorOneByteParsableTest(items=[])

        vector.append(OneByteParsable(1))
        self.assertEqual(vector[0], OneByteParsable(1))
        self.assertNotEqual(vector[0], TwoByteParsable(1))
        self.assertEqual(len(vector), 1)
        self.assertEqual(str(vector), '[0x01]')

        vector.insert(0, TwoByteParsable(0))
        self.assertEqual(vector[0], TwoByteParsable(0))
        self.assertNotEqual(vector[0], OneByteParsable(0))
        self.assertEqual(len(vector), 2)
        self.assertEqual(str(vector), '[0x0000, 0x01]')

        del vector[0]
        self.assertEqual(vector[0], OneByteParsable(1))
        self.assertNotEqual(vector[0], TwoByteParsable(1))
        self.assertEqual(len(vector), 1)
        self.assertEqual(str(vector), '[0x01]')

        vector[0] = TwoByteParsable(0)
        self.assertEqual(vector[0], TwoByteParsable(0))
        self.assertNotEqual(vector[0], OneByteParsable(0))
        self.assertEqual(len(vector), 1)
        self.assertEqual(str(vector), '[0x0000]')


class TestVectorDerived(unittest.TestCase):
    def test_error(self):
        with self.assertRaises(NotEnoughData) as context_manager:
            VectorConsditionalParsableTest.parse_exact_size(b'\x05\x01\x02\x00')
        self.assertEqual(context_manager.exception.bytes_needed, 2)

    def test_parse(self):
        self.assertEqual(
            [0x01, 0x0200],
            list(map(int, VectorConsditionalParsableTest.parse_exact_size(b'\x03\x01\x02\x00')))
        )
        self.assertEqual(
            [0x0200, 0x01],
            list(map(int, VectorConsditionalParsableTest.parse_exact_size(b'\x03\x02\x00\x01')))
        )

    def test_compose(self):
        self.assertEqual(
            b'\x03\x01\x00\x02',
            VectorConsditionalParsableTest([
                OneByteParsable(1),
                TwoByteParsable(2),
            ]).compose()
        )
        self.assertEqual(
            b'\x03\x00\x02\x01',
            VectorConsditionalParsableTest([
                TwoByteParsable(2),
                OneByteParsable(1),
            ]).compose()
        )


class TestOpaque(unittest.TestCase):
    def test_error(self):
        with self.assertRaises(NotEnoughData) as context_manager:
            OpaqueTest.parse_exact_size(b'\x03\x01\x02')
        self.assertEqual(context_manager.exception.bytes_needed, 1)

    def test_parse(self):
        self.assertEqual(
            [1, 2, 3],
            list(OpaqueTest.parse_exact_size(b'\x03\x01\x02\x03'))
        )

    def test_compose(self):
        self.assertEqual(
            b'\x03\x01\x02\x03',
            OpaqueTest([1, 2, 3]).compose()
        )
        self.assertEqual(
            b'\x03\x01\x02\x03',
            OpaqueTest(b'\x01\x02\x03').compose()
        )
        self.assertEqual(
            b'\x03\x01\x02\x03',
            OpaqueTest(bytearray(b'\x01\x02\x03')).compose()
        )


class TestOpaqueEnum(unittest.TestCase):
    def test_error(self):
        with self.assertRaises(InvalidValue) as context_manager:
            OpaqueEnumFactory.parse_exact_size(b'\x0a' + 'δέλτα'.encode('utf-8'))
        self.assertEqual(context_manager.exception.value, 'δέλτα')

    def test_parse(self):
        self.assertEqual(
            OpaqueEnum.ALPHA,
            OpaqueEnumFactory.parse_exact_size(b'\x08' + 'άλφα'.encode('utf-8'))
        )

    def test_compose(self):
        self.assertEqual(
            b'\x0a' + 'γάμμα'.encode('utf-8'),
            OpaqueEnum.GAMMA.compose()  # pylint: disable=no-member
        )

    def test_repr(self):
        self.assertEqual(
            repr(OpaqueEnum.GAMMA),
            'OpaqueEnum.GAMMA'
        )


class TestNByteEnumParsable(unittest.TestCase):
    def test_compose(self):
        composer = ComposerBinary()
        composer.compose_parsable(OneByteEnumComposerTest.ONE)
        self.assertEqual(composer.composed_bytes, b'\x01')

        composer = ComposerBinary()
        composer.compose_parsable(TwoByteEnumComposerTest.TWO)
        self.assertEqual(composer.composed_bytes, b'\x00\x02')

        composer = ComposerBinary()
        composer.compose_parsable(ThreeByteEnumComposerTest.THREE)
        self.assertEqual(composer.composed_bytes, b'\x00\x00\x03')

        composer = ComposerBinary()
        composer.compose_parsable(FourByteEnumComposerTest.FOUR)
        self.assertEqual(composer.composed_bytes, b'\x00\x00\x00\x04')

    def test_parse(self):
        self.assertEqual(
            OneByteEnumParsableTest.parse_exact_size(b'\x01'),
            NByteEnumTest.ONE
        )
        self.assertEqual(
            TwoByteEnumParsableTest.parse_exact_size(b'\x00\x02'),
            NByteEnumTest.TWO
        )
        self.assertEqual(
            ThreeByteEnumParsableTest.parse_exact_size(b'\x00\x00\x03'),
            NByteEnumTest.THREE
        )
        self.assertEqual(
            FourByteEnumParsableTest.parse_exact_size(b'\x00\x00\x00\x04'),
            NByteEnumTest.FOUR
        )


class TestEnumString(unittest.TestCase):
    def test_error(self):
        with self.assertRaises(InvalidValue) as context_manager:
            StringEnum.parse_exact_size(b'four')
        self.assertEqual(context_manager.exception.value, b'four')

        with self.assertRaises(InvalidValue) as context_manager:
            StringEnum.parse_exact_size(b'\xffthree')
        self.assertEqual(context_manager.exception.value, b'\xffthree')

    def test_parse(self):
        self.assertEqual(StringEnum.parse_exact_size(b'one'), StringEnum.ONE)

    def test_compose(self):
        self.assertEqual(StringEnum.ONE.compose(), b'one')


class TestSerializable(unittest.TestCase):
    def test_json(self):
        self.assertEqual(
            SerializableSimpleTypes().as_json(),
            '{' +
            '"bool_value": false, ' +
            '"bytearray_value": "00:01:01:02:03:04:05:06:07:08:09:0A:0B:0C:0D:0E:0F", ' +
            '"float_value": 1.0, ' +
            '"int_value": 1, ' +
            '"none_value": null, ' +
            '"str_value": "string"' +
            '}'
        )
        self.assertEqual(
            SerializableIterables().as_json(),
            '{' +
            '"dict_key": ' +
            '{' +
            '"1": "int", ' +
            '"str": "string", ' +
            '"FIRST": "enum"' +
            '}, ' +
            '"dict_value": ' +
            '{' +
            '"int": 1, ' +
            '"string": "str", ' +
            '"enum": {"FIRST": "1"}' +
            '}, ' +
            '"list_value": ["value"], ' +
            '"tuple_value": ["value"]' +
            '}'
        )
        self.assertEqual(
            SerializableEnums().as_json(),
            '{"param_enum": {"FIRST": {"code": 1}}, "string_enum": {"SECOND": "2"}}'
        )
        self.assertEqual(
            SerializableSingle().as_json(),
            '"single"'
        )
        self.assertEqual(
            SerializableHidden().as_json(),
            '{"visible_value": "value"}'
        )
        self.assertEqual(
            SerializableUnhandled().as_json(),
            '{"complex_number": "(1+2j)"}'
        )
        self.assertEqual(
            SerializableHumanFriendly().as_json(),
            '{' +
            '"human_friendly": "human-friendly", ' +
            '"human_friendly_by_default": "human-friendly-by-default", ' +
            '"non_human_friendly": "non-human-friendly"' +
            '}'
        )
        self.assertEqual(
            SerializableRecursive().as_json(),
            '{' +
            '"json_asdict_object": {"attr_b": "b", "attr_a": "a"}, ' +
            '"json_attr_as_dict": {"attr_b": "b", "attr_a": "a"}, ' +
            '"json_attr_object": {"attr_b": "b", "attr_a": "a"}, ' +
            '"json_crypto_data_hub_enum": "ONE", ' +
            '"json_gradeable": {"vulnerabilities": [{"attack_type": "MITM", "grade": "INSECURE", "named": "NOFS"}]}, ' +
            '"json_object": {"attr_a": "a", "attr_b": "b"}, ' +
            '"json_serializable_hidden": {"visible_value": "value"}, ' +
            '"json_serializable_in_dict": {"key1": {"visible_value": "value"}, "key2": "single"}, ' +
            '"json_serializable_in_list": [{"visible_value": "value"}, "single"], ' +
            '"json_serializable_in_tuple": [{"visible_value": "value"}, "single"], ' +
            '"json_serializable_single": "single"' +
            '}'
        )
        self.assertEqual(json.dumps(TestObject()), '{}')
        self.assertEqual(
            json.dumps(EnumStringValue.ONE),
            '{"ONE": "one"}'
        )

    def test_markdown(self):
        self.assertEqual(
            SerializableSimpleTypes().as_markdown(),
            '\n'.join([
                '* Bool Value: no',
                '* Bytearray Value: 00:01:01:02:03:04:05:06:07:08:09:0A:0B:0C:0D:0E:0F',
                '* Float Value: 1.0',
                '* Int Value: 1',
                '* None Value: n/a',
                '* Str Value: string',
                ''
            ])
        )
        self.assertEqual(
            SerializableIterables().as_markdown(),
            '\n'.join([
                '* Dict Key:',
                '    * 1: int',
                '    * Str: string',
                '    * FIRST: enum',
                '* Dict Value:',
                '    * Int: 1',
                '    * String: str',
                '    * Enum: FIRST',
                '* List Value:',
                '    1. value',
                '* Tuple Value:',
                '    1. value',
                ''
            ])
        )
        self.assertEqual(
            SerializableEnums().as_markdown(),
            '\n'.join([
                '* Param Enum: 1',
                '* String Enum: SECOND',
                ''
            ])
        )
        self.assertEqual(
            SerializableHidden().as_markdown(),
            '* Visible Value: value\n'
        )
        self.assertEqual(
            SerializableUnhandled().as_markdown(),
            '* Complex Number: (1+2j)\n'
        )
        self.assertEqual(
            SerializableHumanReadable().as_markdown(),
            '* Human Readable Name 2: value 2\n'
            '* Human Readable Name 1: value 1\n'
        )
        self.assertEqual(
            SerializableHumanFriendly().as_markdown(),
            '* Human Friendly: human-friendly\n'
            '* Human Friendly By Default: human-friendly-by-default\n'
        )
        self.assertEqual(
            SerializableAttributeOrder().as_markdown(),
            '\n'.join([
                '* Attr B: b',
                '* Attr A: a'
            ]) + '\n'
        )
        self.assertEqual(
            SerializableEmptyValues().as_markdown(),
            '\n'.join([
                '* Dict: -',
                '* List: -',
                '* Tuple: -',
                '* Value: n/a',
                '',
            ])
        )
        self.assertEqual(
            SerializableRecursive().as_markdown(),
            '\n'.join([
                '* Json Asdict Object:',
                '    * Attr B: b',
                '    * Attr A: a',
                '* Json Attr As Dict:',
                '    * Attr B: b',
                '    * Attr A: a',
                '* Json Attr Object:',
                '    * Attr B: b',
                '    * Attr A: a',
                '* Json Crypto Data Hub Enum: one',
                '* Json Gradeable: value',
                '* Json Object:',
                '    * Attr A: a',
                '    * Attr B: b',
                '* Json Serializable Hidden:',
                '    * Visible Value: value',
                '* Json Serializable In Dict:',
                '    * Key1:',
                '        * Visible Value: value',
                '    * Key2: single',
                '* Json Serializable In List:',
                '    1.',
                '        * Visible Value: value',
                '    2. single',
                '* Json Serializable In Tuple:',
                '    1.',
                '        * Visible Value: value',
                '    2. single',
                '* Json Serializable Single: single',
                '',
            ])
        )

        Serializable.post_text_encoder = SerializableUpperCaseEncoder()
        self.assertEqual(
            SerializableRecursive().as_markdown(),
            '\n'.join([
                '* Json Asdict Object:',
                '    * Attr B: B',
                '    * Attr A: A',
                '* Json Attr As Dict:',
                '    * Attr B: B',
                '    * Attr A: A',
                '* Json Attr Object:',
                '    * Attr B: B',
                '    * Attr A: A',
                '* Json Crypto Data Hub Enum: ONE',
                '* Json Gradeable: VALUE',
                '* Json Object:',
                '    * Attr A: A',
                '    * Attr B: B',
                '* Json Serializable Hidden:',
                '    * Visible Value: VALUE',
                '* Json Serializable In Dict:',
                '    * Key1:',
                '        * Visible Value: VALUE',
                '    * Key2: SINGLE',
                '* Json Serializable In List:',
                '    1.',
                '        * Visible Value: VALUE',
                '    2. SINGLE',
                '* Json Serializable In Tuple:',
                '    1.',
                '        * Visible Value: VALUE',
                '    2. SINGLE',
                '* Json Serializable Single: SINGLE',
                '',
            ])
        )
        Serializable.post_text_encoder = SerializableTextEncoder()


class TestListParsable(unittest.TestCase):
    def test_error(self):
        with self.assertRaises(NotEnoughData) as context_manager:
            ListParsableTest.parse_exact_size(b'')
        self.assertEqual(context_manager.exception.bytes_needed, 2)

        with self.assertRaises(InvalidValue) as context_manager:
            ListParsableTest.parse_exact_size(b'test')
        self.assertEqual(context_manager.exception.value, b'test')

        with self.assertRaises(InvalidValue) as context_manager:
            ListParsableTest.parse_exact_size(b'nottest\r\n\r\n')
        self.assertEqual(context_manager.exception.value, b'nottest\r\n\r\n')

        with self.assertRaises(InvalidValue) as context_manager:
            ListParsableTest.parse_exact_size(b'test\r\n\r\ntest\r\n\r\n')
        self.assertEqual(context_manager.exception.value, b'test\r\n\r\ntest\r\n\r\n')

    def test_parse(self):
        list_parsable = ListParsableTest.parse_exact_size(b'\r\n')
        self.assertEqual(list_parsable, ListParsableTest([]))

        list_parsable = ListParsableTest.parse_exact_size(b'test\r\n\r\n')
        self.assertEqual(list_parsable, ListParsableTest([AlwaysTestStringComposer(), ]))

        list_parsable = ListParsableTest.parse_exact_size(b'test\r\ntest\r\n\r\n')
        self.assertEqual(list_parsable, ListParsableTest([AlwaysTestStringComposer(), AlwaysTestStringComposer()]))

    def test_compose(self):
        self.assertEqual(
            ListParsableTest([]).compose(),
            b'\r\n'
        )
        self.assertEqual(
            ListParsableTest([AlwaysTestStringComposer(), ]).compose(),
            b'test\r\n\r\n'
        )
        self.assertEqual(
            ListParsableTest([AlwaysTestStringComposer(), AlwaysTestStringComposer(), ]).compose(),
            b'test\r\ntest\r\n\r\n'
        )


class TestVariantParsable(unittest.TestCase):
    def test_error(self):
        with self.assertRaises(TooMuchData) as context_manager:
            VariantParsableTest.parse_exact_size(b'aaaa')
        self.assertEqual(context_manager.exception.bytes_needed, 1)

    def test_parse(self):
        parsable = bytearray(b'a')
        self.assertEqual(VariantParsableTest.parse_mutable(parsable), StringEnumA.A)
        self.assertEqual(parsable, b'')

        parsable = bytearray(b'aa')
        self.assertEqual(VariantParsableTest.parse_mutable(parsable), StringEnumA.A)
        self.assertEqual(parsable, b'a')

        parsable = bytearray(b'aaa')
        self.assertEqual(VariantParsableTest.parse_mutable(parsable), StringEnumA.A)
        self.assertEqual(parsable, b'aa')


class TestVariantParsableExact(unittest.TestCase):
    def test_error(self):
        with self.assertRaises(InvalidValue) as context_manager:
            VariantParsableExactTest.parse_exact_size(b'aaaa')
        self.assertEqual(context_manager.exception.value, b'aaaa')

    def test_parse(self):
        parsable = bytearray(b'a')
        self.assertEqual(VariantParsableExactTest.parse_mutable(parsable), StringEnumA.A)
        self.assertEqual(parsable, b'')

        parsable = bytearray(b'aa')
        self.assertEqual(VariantParsableExactTest.parse_mutable(parsable), StringEnumAA.AA)
        self.assertEqual(parsable, b'')

        parsable = bytearray(b'aaa')
        self.assertEqual(VariantParsableExactTest.parse_mutable(parsable), StringEnumAAA.AAA)
        self.assertEqual(parsable, b'')


class TestNumericRangeParsable(unittest.TestCase):
    def test_error(self):
        with self.assertRaises(InvalidValue) as context_manager:
            NumericRangeParsableTest.parse_exact_size(b'\x00')
        self.assertEqual(context_manager.exception.value, 0x00)

        with self.assertRaises(InvalidValue) as context_manager:
            NumericRangeParsableTest.parse_exact_size(b'\xff')
        self.assertEqual(context_manager.exception.value, 0xff)

    def test_parse(self):
        self.assertEqual(NumericRangeParsableTest.parse_exact_size(b'\x01'), NumericRangeParsableTest(1))
        self.assertEqual(NumericRangeParsableTest(1).compose(), b'\x01')

    def test_str(self):
        self.assertEqual(str(NumericRangeParsableTest(1)), '1')

    def test_as_markdown(self):
        self.assertEqual(NumericRangeParsableTest(1).as_markdown(), '1')



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import abc
import collections
import enum
import hashlib
import random

import attr

from cryptodatahub.common.exception import InvalidValue
from cryptodatahub.ssh.algorithm import (
    SshKexAlgorithm,
    SshHostKeyAlgorithm,
    SshEncryptionAlgorithm,
    SshMacAlgorithm,
    SshCompressionAlgorithm,
)

from cryptoparser.common.base import (
    VariantParsable,
    VectorParamString,
    VectorString,
)
from cryptoparser.common.classes import LanguageTag
from cryptoparser.common.exception import InvalidType, TooMuchData
from cryptoparser.common.parse import ParsableBase, ParserBinary, ComposerBinary, ParserText, ComposerText
from cryptoparser.common.utils import bytes_to_hex_string

from cryptoparser.ssh.key import SshPublicKeyBase, SshHostPublicKeyVariant
from cryptoparser.ssh.version import (
    SshProtocolVersion,
    SshSoftwareVersionBase,
    SshSoftwareVersionParsedVariant,
    SshSoftwareVersionUnparsed,
)


class SshMessageCode(enum.IntEnum):
    DISCONNECT = 0x1
    IGNORE = 0x2
    UNIMPLEMENTED = 0x3
    DEBUG = 0x4
    SERVICE_REQUEST = 0x5
    SERVICE_ACCEPT = 0x6
    KEXINIT = 0x14
    NEWKEYS = 0x15
    DH_KEX_INIT = 0x1e
    DH_KEX_REPLY = 0x1f
    DH_GEX_GROUP = 0x1f
    DH_GEX_INIT = 0x20
    DH_GEX_REPLY = 0x21
    DH_GEX_REQUEST = 0x22


class SshMessageBase(ParsableBase):
    @classmethod
    @abc.abstractmethod
    def get_message_code(cls):
        raise NotImplementedError()

    @classmethod
    def _parse_header(cls, parsable):
        parser = ParserBinary(parsable)

        parser.parse_numeric('message_code', 1, SshMessageCode)

        if parser['message_code'] != cls.get_message_code():
            raise InvalidType()

        return parser

    @classmethod
    def _compose_header(cls):
        composer = ComposerBinary()

        composer.compose_numeric(cls.get_message_code(), 1)

        return composer


@attr.s
class SshProtocolMessage(ParsableBase):
    protocol_version = attr.ib(validator=attr.validators.instance_of(SshProtocolVersion))
    software_version = attr.ib(validator=attr.validators.instance_of(SshSoftwareVersionBase))
    comment = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(str)), default=None)

    @comment.validator
    def comment_validator(self, _, value):  # pylint: disable=no-self-use
        if value is not None:
            if '\r' in value or '\n' in value:
                raise InvalidValue(value, SshProtocolMessage, 'comment')
            try:
                value.encode('ascii')
            except UnicodeEncodeError as e:
                raise InvalidValue(value, SshProtocolMessage, 'comment') from e

    @classmethod
    def _parse(cls, parsable):
        parser = ParserText(parsable)

        parser.parse_string_by_length('protocol', min_length=3, max_length=3)
        if parser['protocol'] != 'SSH':
            raise InvalidValue(parser['protocol'], SshProtocolMessage, 'protocol')

        parser.parse_string('separator', '-')
        parser.parse_parsable('protocol_version', SshProtocolVersion)
        parser.parse_string('separator', '-')

        parser.parse_string_until_separator('software_version_and_comment', '\n')
        software_version_and_comment = parser['software_version_and_comment'].split(' ')

        if software_version_and_comment[-1][-1] == '\r':
            software_version_and_comment[-1] = software_version_and_comment[-1][:-1]

        software_version_parser = ParserText(software_version_and_comment[0].encode('ascii'))
        try:
            software_version_parser.parse_parsable('value', SshSoftwareVersionParsedVariant)
        except InvalidValue:
            software_version_parser.parse_parsable('value', SshSoftwareVersionUnparsed)

        if len(software_version_and_comment) > 1:
            comment = ' '.join(software_version_and_comment[1:])
        else:
            comment = None
        parser.parse_separator('\n')

        if parser.parsed_length > 255:
            raise TooMuchData(parser.parsed_length - 255)

        return SshProtocolMessage(
            parser['protocol_version'],
            software_version_parser['value'],
            comment
        ), parser.parsed_length

    def compose(self):
        composer = ComposerText()

        composer.compose_string('SSH')
        composer.compose_separator('-')
        composer.compose_parsable(self.protocol_version)
        composer.compose_separator('-')
        composer.compose_string(self.software_version)
        if self.comment is not None:
            composer.compose_separator(' ')
            composer.compose_string(self.comment)
        composer.compose_separator('\r\n')

        return composer.composed


class SshAlgorithmVector(VectorString):
    @classmethod
    def get_param(cls):
        return VectorParamString(
            min_byte_num=0,
            max_byte_num=2 ** 32 - 1,
            separator=',',
            item_class=cls.get_item_class(),
            fallback_class=str
        )

    @classmethod
    @abc.abstractmethod
    def get_item_class(cls):
        raise NotImplementedError()


class SshKexAlgorithmVector(SshAlgorithmVector):
    @classmethod
    def get_item_class(cls):
        return SshKexAlgorithm


class SshHostKeyAlgorithmVector(SshAlgorithmVector):
    @classmethod
    def get_item_class(cls):
        return SshHostKeyAlgorithm


class SshEncryptionAlgorithmVector(SshAlgorithmVector):
    @classmethod
    def get_item_class(cls):
        return SshEncryptionAlgorithm


class SshMacAlgorithmVector(SshAlgorithmVector):
    @classmethod
    def get_item_class(cls):
        return SshMacAlgorithm


class SshCompressionAlgorithmVector(SshAlgorithmVector):
    @classmethod
    def get_item_class(cls):
        return SshCompressionAlgorithm


class VectorParamSshLanguage(VectorParamString):
    def __init__(self):
        super().__init__(
            min_byte_num=0,
            max_byte_num=2 ** 32 - 1,
            separator=',',
            item_class=LanguageTag,
            fallback_class=None,
        )

    def get_item_size(self, item):
        return len(item.compose())


class SshLanguageVector(VectorString):
    @classmethod
    def get_param(cls):
        return VectorParamSshLanguage()


@attr.s
class SshKeyExchangeInit(SshMessageBase):  # pylint: disable=too-many-instance-attributes
    kex_algorithms = attr.ib(
        converter=SshKexAlgorithmVector,
        validator=attr.validators.instance_of(SshKexAlgorithmVector)
    )
    host_key_algorithms = attr.ib(
        converter=SshHostKeyAlgorithmVector,
        validator=attr.validators.instance_of(SshHostKeyAlgorithmVector)
    )
    encryption_algorithms_client_to_server = attr.ib(
        converter=SshEncryptionAlgorithmVector,
        validator=attr.validators.instance_of(SshEncryptionAlgorithmVector)
    )
    encryption_algorithms_server_to_client = attr.ib(
        converter=SshEncryptionAlgorithmVector,
        validator=attr.validators.instance_of(SshEncryptionAlgorithmVector)
    )
    mac_algorithms_client_to_server = attr.ib(
        converter=SshMacAlgorithmVector,
        validator=attr.validators.instance_of(SshMacAlgorithmVector)
    )
    mac_algorithms_server_to_client = attr.ib(
        converter=SshMacAlgorithmVector,
        validator=attr.validators.instance_of(SshMacAlgorithmVector)
    )
    compression_algorithms_client_to_server = attr.ib(
        converter=SshCompressionAlgorithmVector,
        validator=attr.validators.instance_of(SshCompressionAlgorithmVector)
    )
    compression_algorithms_server_to_client = attr.ib(
        converter=SshCompressionAlgorithmVector,
        validator=attr.validators.instance_of(SshCompressionAlgorithmVector)
    )
    languages_client_to_server = attr.ib(
        converter=SshLanguageVector,
        validator=attr.validators.instance_of(SshLanguageVector), default=()
    )
    languages_server_to_client = attr.ib(
        converter=SshLanguageVector,
        validator=attr.validators.instance_of(SshLanguageVector), default=()
    )
    first_kex_packet_follows = attr.ib(validator=attr.validators.instance_of(int), default=0)
    cookie = attr.ib(
        validator=attr.validators.instance_of((bytearray, bytes)),
        default=bytearray.fromhex(f'{random.getrandbits(128):16x}'.zfill(32))
    )
    reserved = attr.ib(validator=attr.validators.instance_of(int), default=0x00000000)

    @classmethod
    def _get_cipher_attributes(cls):
        for attribute in attr.fields(cls):
            if (attribute.validator and
                    isinstance(attribute.validator.type, type) and
                    issubclass(attribute.validator.type, ParsableBase)):
                yield attribute

    @classmethod
    def _parse(cls, parsable):
        header_parser = cls._parse_header(parsable)

        body_parser = ParserBinary(parsable[header_parser.parsed_length:])

        body_parser.parse_raw('cookie', 16)
        for attribute in cls._get_cipher_attributes():
            body_parser.parse_parsable(attribute.name, attribute.validator.type)

        body_parser.parse_numeric('first_kex_packet_follows', 1, bool)
        body_parser.parse_numeric('reserved', 4)

        return SshKeyExchangeInit(**dict(body_parser)), header_parser.parsed_length + body_parser.parsed_length

    def compose(self):
        composer = self._compose_header()

        composer.compose_raw(self.cookie)

        for attribute in self._get_cipher_attributes():
            composer.compose_parsable(getattr(self, attribute.name))

        composer.compose_numeric(1 if self.first_kex_packet_follows else 0, 1)
        composer.compose_numeric(self.reserved, 4)

        return composer.composed

    @classmethod
    def get_message_code(cls):
        return SshMessageCode.KEXINIT

    @staticmethod
    def _hassh(algorithm_vectors):
        hassh_text = ';'.join([
            ','.join([
                algorithm if isinstance(algorithm, str) else algorithm.value.code
                for algorithm in algorithms
            ])
            for algorithms in algorithm_vectors
        ])

        message = hashlib.md5()
        message.update(hassh_text.encode('ascii'))

        return bytes_to_hex_string(message.digest(), lowercase=True)

    @property
    def hassh(self):
        return self._hassh([
            self.kex_algorithms,
            self.encryption_algorithms_client_to_server,
            self.mac_algorithms_client_to_server,
            self.compression_algorithms_client_to_server,
        ])

    @property
    def hassh_server(self):
        return self._hassh([
            self.kex_algorithms,
            self.encryption_algorithms_server_to_client,
            self.mac_algorithms_server_to_client,
            self.compression_algorithms_server_to_client,
        ])


class SshReasonCode(enum.IntEnum):
    HOST_NOT_ALLOWED_TO_CONNECT = 0x01
    PROTOCOL_ERROR = 0x02
    KEY_EXCHANGE_FAILED = 0x03
    RESERVED = 0x04
    MAC_ERROR = 0x05
    COMPRESSION_ERROR = 0x06
    SERVICE_NOT_AVAILABLE = 0x07
    PROTOCOL_VERSION_NOT_SUPPORTED = 0x08
    HOST_KEY_NOT_VERIFIABLE = 0x09
    CONNECTION_LOST = 0x0a
    BY_APPLICATION = 0x0b
    TOO_MANY_CONNECTIONS = 0x0c
    AUTH_CANCELLED_BY_USER = 0x0d
    NO_MORE_AUTH_METHODS_AVAILABLE = 0x0e
    ILLEGAL_USER_NAME = 0x0f


@attr.s
class SshDisconnectMessage(SshMessageBase):
    reason = attr.ib(validator=attr.validators.instance_of(SshReasonCode))
    description = attr.ib(
        converter=str,
        validator=attr.validators.instance_of(str)
    )
    language = attr.ib(
        default='US',
        validator=attr.validators.instance_of(str)
    )

    @classmethod
    def get_message_code(cls):
        return SshMessageCode.DISCONNECT

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)

        parser.parse_numeric('reason', 4, SshReasonCode)
        parser.parse_string('description', 4, 'utf-8')
        parser.parse_string('language', 4, 'ascii')

        return SshDisconnectMessage(parser['reason'], parser['description'], parser['language']), parser.parsed_length

    def compose(self):
        composer = self._compose_header()

        composer.compose_numeric(self.reason.value, 4)
        composer.compose_string(self.description, 'utf-8', 4)
        composer.compose_string(self.language, 'ascii', 4)

        return composer.composed


@attr.s
class SshUnimplementedMessage(SshMessageBase):
    sequence_number = attr.ib(validator=attr.validators.instance_of(int))

    @classmethod
    def get_message_code(cls):
        return SshMessageCode.UNIMPLEMENTED

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)

        parser.parse_numeric('sequence_number', 4)

        return SshUnimplementedMessage(parser['sequence_number']), parser.parsed_length

    def compose(self):
        composer = self._compose_header()

        composer.compose_numeric(self.sequence_number, 4)

        return composer.composed


@attr.s
class SshDHKeyExchangeInitBase(SshMessageBase):
    ephemeral_public_key = attr.ib(validator=attr.validators.instance_of((bytes, bytearray)))

    @classmethod
    @abc.abstractmethod
    def get_message_code(cls):
        raise NotImplementedError()

    @classmethod
    def _parse(cls, parsable):
        header_parser = cls._parse_header(parsable)

        body_parser = ParserBinary(parsable[header_parser.parsed_length:])
        body_parser.parse_bytes('ephemeral_public_key', 4)

        return cls(
            body_parser['ephemeral_public_key']
        ), header_parser.parsed_length + body_parser.parsed_length

    def compose(self):
        composer = self._compose_header()

        composer.compose_bytes(self.ephemeral_public_key, 4)

        return composer.composed


class SshDHKeyExchangeInit(SshDHKeyExchangeInitBase):
    @classmethod
    def get_message_code(cls):
        return SshMessageCode.DH_KEX_INIT


class SshDHGroupExchangeInit(SshDHKeyExchangeInitBase):
    @classmethod
    def get_message_code(cls):
        return SshMessageCode.DH_GEX_INIT


@attr.s
class SshDHKeyExchangeReplyBase(SshMessageBase):
    host_public_key = attr.ib(validator=attr.validators.instance_of(SshPublicKeyBase))
    ephemeral_public_key = attr.ib(validator=attr.validators.instance_of((bytes, bytearray)))
    signature = attr.ib(validator=attr.validators.instance_of((bytes, bytearray)))

    @classmethod
    @abc.abstractmethod
    def get_message_code(cls):
        raise NotImplementedError()

    @staticmethod
    def _parse_ssh_key(parser):
        parser.parse_parsable('host_public_key', SshHostPublicKeyVariant, 4)
        parser.parse_bytes('ephemeral_public_key', 4)
        parser.parse_bytes('signature', 4)

        return parser['host_public_key'], parser['ephemeral_public_key'], parser['signature']

    @staticmethod
    def _compose_ssh_key(composer, host_public_key, ephemeral_public_key_bytes, signature_bytes):
        composer.compose_parsable(host_public_key, 4)
        composer.compose_bytes(ephemeral_public_key_bytes, 4)
        composer.compose_bytes(signature_bytes, 4)

    @classmethod
    def _parse(cls, parsable):

        parser = cls._parse_header(parsable)

        host_public_key, ephemeral_public_key, signature = cls._parse_ssh_key(parser)

        return cls(
            host_public_key,
            ephemeral_public_key,
            signature,
        ), parser.parsed_length

    def compose(self):
        composer = self._compose_header()

        self._compose_ssh_key(composer, self.host_public_key, self.ephemeral_public_key, self.signature)

        return composer.composed


class SshDHKeyExchangeReply(SshDHKeyExchangeReplyBase):
    @classmethod
    def get_message_code(cls):
        return SshMessageCode.DH_KEX_REPLY


class SshDHGroupExchangeReply(SshDHKeyExchangeReplyBase):
    @classmethod
    def get_message_code(cls):
        return SshMessageCode.DH_GEX_REPLY


@attr.s
class SshDHGroupExchangeRequest(SshMessageBase):
    gex_min = attr.ib(validator=attr.validators.instance_of(int))
    gex_number = attr.ib(validator=attr.validators.instance_of(int))
    gex_max = attr.ib(validator=attr.validators.instance_of(int))

    @classmethod
    def get_message_code(cls):
        return SshMessageCode.DH_GEX_REQUEST

    @classmethod
    def _parse(cls, parsable):
        header_parser = cls._parse_header(parsable)

        body_parser = ParserBinary(parsable[header_parser.parsed_length:])
        body_parser.parse_numeric('gex_min', 4)
        body_parser.parse_numeric('gex_number', 4)
        body_parser.parse_numeric('gex_max', 4)

        return SshDHGroupExchangeRequest(
            body_parser['gex_min'],
            body_parser['gex_number'],
            body_parser['gex_max'],
        ), header_parser.parsed_length + body_parser.parsed_length

    def compose(self):
        composer = self._compose_header()

        composer.compose_numeric(self.gex_min, 4)
        composer.compose_numeric(self.gex_number, 4)
        composer.compose_numeric(self.gex_max, 4)

        return composer.composed


@attr.s
class SshDHGroupExchangeGroup(SshMessageBase):
    p = attr.ib(validator=attr.validators.instance_of((bytes, bytearray)))
    g = attr.ib(validator=attr.validators.instance_of((bytes, bytearray)))

    @classmethod
    def get_message_code(cls):
        return SshMessageCode.DH_GEX_GROUP

    @classmethod
    def _parse(cls, parsable):
        header_parser = cls._parse_header(parsable)

        body_parser = ParserBinary(parsable[header_parser.parsed_length:])
        body_parser.parse_bytes('p', 4)
        body_parser.parse_bytes('g', 4)

        return SshDHGroupExchangeGroup(
            body_parser['p'],
            body_parser['g'],
        ), header_parser.parsed_length + body_parser.parsed_length

    def compose(self):
        composer = self._compose_header()

        composer.compose_bytes(self.p, 4)
        composer.compose_bytes(self.g, 4)

        return composer.composed


@attr.s
class SshNewKeys(SshMessageBase):
    @classmethod
    def get_message_code(cls):
        return SshMessageCode.NEWKEYS

    @classmethod
    def _parse(cls, parsable):
        header_parser = cls._parse_header(parsable)

        return SshNewKeys(), header_parser.parsed_length

    def compose(self):
        composer = self._compose_header()

        return composer.composed


class SshMessageVariantInit(VariantParsable):
    _VARIANTS = collections.OrderedDict([
        (SshMessageCode.DISCONNECT, (SshDisconnectMessage, )),
        (SshMessageCode.UNIMPLEMENTED, (SshUnimplementedMessage, )),
        (SshMessageCode.KEXINIT, (SshKeyExchangeInit, )),
    ])

    @classmethod
    def _get_variants(cls):
        return cls._VARIANTS


class SshMessageVariantKexDH(VariantParsable):
    _VARIANTS = collections.OrderedDict([
        (SshMessageCode.DISCONNECT, (SshDisconnectMessage, )),
        (SshMessageCode.KEXINIT, (SshKeyExchangeInit, )),
        (SshMessageCode.UNIMPLEMENTED, (SshUnimplementedMessage, )),
        (SshMessageCode.DH_KEX_INIT, (SshDHKeyExchangeInit, )),
        (SshMessageCode.DH_KEX_REPLY, (SshDHKeyExchangeReply, )),
        (SshMessageCode.NEWKEYS, (SshNewKeys, )),
    ])

    @classmethod
    def _get_variants(cls):
        return cls._VARIANTS


class SshMessageVariantKexDHGroup(VariantParsable):
    _VARIANTS = collections.OrderedDict([
        (SshMessageCode.DISCONNECT, (SshDisconnectMessage, )),
        (SshMessageCode.KEXINIT, (SshKeyExchangeInit, )),
        (SshMessageCode.UNIMPLEMENTED, (SshUnimplementedMessage, )),
        (SshMessageCode.DH_GEX_REQUEST, (SshDHGroupExchangeRequest, )),
        (SshMessageCode.DH_GEX_GROUP, (SshDHGroupExchangeGroup, )),
        (SshMessageCode.DH_GEX_INIT, (SshDHGroupExchangeInit, )),
        (SshMessageCode.DH_GEX_REPLY, (SshDHGroupExchangeReply, )),
        (SshMessageCode.NEWKEYS, (SshNewKeys, )),
    ])

    @classmethod
    def _get_variants(cls):
        return cls._VARIANTS



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

"""
Test ISAKMP header parsing and composition.
"""

import collections
import unittest

from test.ike.classes import Ikev2PayloadBaseTest

from cryptodatahub.common.exception import InvalidValue
from cryptodatahub.ike.algorithm import (
    Ikev1ExchangeType,
    Ikev1PayloadType,
    Ikev2DiffieHellmanGroup,
    Ikev2ExchangeType,
    Ikev2PayloadType,
)

from cryptoparser.common.exception import InvalidType, NotEnoughData
from cryptoparser.ike.ikev1 import Ikev1PayloadKeyExchange, Ikev1PayloadNonce
from cryptoparser.ike.ikev2 import Ikev2PayloadNonce, Ikev2PayloadKeyExchange
from cryptoparser.ike.isakmp import IsakmpFlags, IsakmpMessage
from cryptoparser.ike.version import IsakmpProtocolVersion, IsakmpVersion


class TestISAKMPHeader(unittest.TestCase):
    def setUp(self):
        self.header_dict = collections.OrderedDict([
            ('initiator_cookie', b'\x00' * 8),
            ('responder_cookie', b'\x00' * 8),
            ('next_payload', b'\x00'),  # NONE
            ('protocol_version', b'\x11'),  # ISAKMP v1.1
            ('exchange_type', b'\x01'),  # BASE
            ('flags', b'\x00'),
            ('message_id', b'\x00' * 4),
            ('length', b'\x00\x00\x00\x1c'),  # 28 bytes
        ])
        self.header_bytes = b''.join(self.header_dict.values())
        self.header = IsakmpMessage(
            version=IsakmpProtocolVersion(IsakmpVersion.V1, 1),
            initiator_spi=0,
            responder_spi=0,
            exchange_type=Ikev1ExchangeType.BASE,
            flags=set(),
            message_id=0,
            payloads=[]
        )

    def test_error(self):
        with self.assertRaises(NotEnoughData) as context_manager:
            IsakmpMessage.parse_exact_size(b'')
        self.assertEqual(context_manager.exception.bytes_needed, IsakmpMessage.HEADER_SIZE)

        with self.assertRaises(InvalidType):
            IsakmpMessage.parse_exact_size(b'\x00' * (IsakmpMessage.HEADER_SIZE - 1) + b'\xff')

    def test_parse(self):
        header = IsakmpMessage.parse_exact_size(self.header_bytes)
        self.assertEqual(header.initiator_spi, 0)
        self.assertEqual(header.responder_spi, 0)
        self.assertEqual(header.version, IsakmpProtocolVersion(IsakmpVersion.V1, 1))
        self.assertEqual(header.exchange_type, Ikev1ExchangeType.BASE)
        self.assertEqual(header.flags, set())
        self.assertEqual(header.message_id, 0)

    def test_compose(self):
        self.assertEqual(self.header.compose(), self.header_bytes)

    def test_flags(self):
        header_with_flags = IsakmpMessage(
            version=IsakmpProtocolVersion(IsakmpVersion.V1, 1),
            initiator_spi=0,
            responder_spi=0,
            exchange_type=Ikev1ExchangeType.BASE,
            flags={IsakmpFlags.ENCRYPTION, IsakmpFlags.COMMIT},
            message_id=0,
            payloads=[]
        )
        header_bytes_with_flags = bytearray(self.header_bytes)
        header_bytes_with_flags[19] = 0x03  # Set ENCRYPTION and COMMIT flags
        self.assertEqual(header_with_flags.compose(), bytes(header_bytes_with_flags))

    def test_ikev1_compose_payload_type(self):
        payload = Ikev1PayloadNonce(nonce_data=b'A' * 16)
        message = IsakmpMessage(
            version=IsakmpProtocolVersion(IsakmpVersion.V1, 0),
            initiator_spi=0,
            responder_spi=0,
            exchange_type=Ikev1ExchangeType.INFORMATIONAL,
            flags=set(),
            message_id=0,
            payloads=[payload],
        )
        composed = message.compose()
        self.assertEqual(composed[16], Ikev1PayloadType.NONCE.value.code)

    def test_ikev2_parsing(self):
        header_dict_v2 = collections.OrderedDict([
            ('initiator_cookie', b'\x00' * 8),
            ('responder_cookie', b'\x00' * 8),
            ('next_payload', b'\x00'),  # NONE
            ('protocol_version', b'\x20'),  # ISAKMP v2.0
            ('exchange_type', b'\x22'),  # IKE_SA_INIT
            ('flags', b'\x00'),
            ('message_id', b'\x00' * 4),
            ('length', b'\x00\x00\x00\x1c'),  # 28 bytes
        ])
        header_bytes_v2 = b''.join(header_dict_v2.values())
        header = IsakmpMessage.parse_exact_size(header_bytes_v2)
        self.assertEqual(header.version, IsakmpProtocolVersion(IsakmpVersion.V2, 0))
        self.assertEqual(header.exchange_type, Ikev2ExchangeType.IKE_SA_INIT)

    def test_invalid_payload_type_ikev1(self):
        header_dict = collections.OrderedDict([
            ('initiator_cookie', b'\x00' * 8),
            ('responder_cookie', b'\x00' * 8),
            ('next_payload', b'\xff'),  # Invalid payload type
            ('protocol_version', b'\x11'),  # ISAKMP v1.1
            ('exchange_type', b'\x01'),  # BASE
            ('flags', b'\x00'),
            ('message_id', b'\x00' * 4),
            ('length', b'\x00\x00\x00\x1c'),  # 28 bytes
        ])
        header_bytes = b''.join(header_dict.values())
        with self.assertRaises(InvalidValue):
            IsakmpMessage.parse_exact_size(header_bytes)

    def test_ikev2_compose_with_empty_payloads(self):
        header_v2 = IsakmpMessage(
            version=IsakmpProtocolVersion(IsakmpVersion.V2, 0),
            initiator_spi=0,
            responder_spi=0,
            exchange_type=Ikev2ExchangeType.IKE_SA_INIT,
            flags=set(),
            message_id=0,
            payloads=[]
        )

        with self.assertRaises(IndexError):
            header_v2.compose()

    def test_payload_parsing_with_invalid_next_payload(self):
        header_dict = collections.OrderedDict([
            ('initiator_cookie', b'\x00' * 8),
            ('responder_cookie', b'\x00' * 8),
            ('next_payload', b'\x01'),  # SECURITY_ASSOCIATION
            ('protocol_version', b'\x11'),  # ISAKMP v1.1
            ('exchange_type', b'\x01'),  # BASE
            ('flags', b'\x00'),
            ('message_id', b'\x00' * 4),
            ('length', b'\x00\x00\x00\x20'),  # 32 bytes (header + 4 bytes payload)
        ])
        payload_data = b'\x00\x00\x00\x04'  # Invalid/minimal payload
        header_bytes = b''.join(header_dict.values()) + payload_data
        with self.assertRaises((NotEnoughData, TypeError, AttributeError)):
            IsakmpMessage.parse_exact_size(header_bytes)

    def test_payload_composition_loop(self):
        payload = Ikev2PayloadBaseTest(
            flags=set(),
            test_data=b'\x00\x01\x02\x03'
        )
        message = IsakmpMessage(
            version=IsakmpProtocolVersion(IsakmpVersion.V2, 0),
            initiator_spi=0,
            responder_spi=0,
            exchange_type=Ikev2ExchangeType.IKE_SA_INIT,
            flags=set(),
            message_id=0,
            payloads=[payload]
        )
        message_bytes = message.compose()
        payload_bytes = payload.compose()
        self.assertEqual(message_bytes[-len(payload_bytes):], payload_bytes)

    def test_payload_parsing_loop(self):
        header_dict = collections.OrderedDict([
            ('initiator_cookie', b'\x00' * 8),
            ('responder_cookie', b'\x00' * 8),
            ('next_payload', b'\x28'),  # NONCE payload type (0x28 = 40 decimal)
            ('protocol_version', b'\x20'),  # ISAKMP v2.0
            ('exchange_type', b'\x22'),  # IKE_SA_INIT
            ('flags', b'\x00'),
            ('message_id', b'\x00' * 4),
            ('length', b'\x00\x00\x00\x30'),  # 48 bytes total (28 header + 20 payload)
        ])
        nonce_payload = (
            b'\x00' +        # Next payload = NONE (0x00)
            b'\x00' +        # Flags = 0 (no critical bit)
            b'\x00\x14' +    # Payload length = 20 bytes (header 4 + data 16)
            b'A' * 16       # 16 bytes of nonce data (minimum required for NONCE)
        )
        header_bytes = b''.join(header_dict.values()) + nonce_payload
        message = IsakmpMessage.parse_exact_size(header_bytes)
        self.assertEqual(len(message.payloads), 1)
        self.assertEqual(message.payloads[0].get_payload_type(), Ikev2PayloadType.NONCE)

    def test_invalid_payload_type_ikev2(self):
        header_dict = collections.OrderedDict([
            ('initiator_cookie', b'\x00' * 8),
            ('responder_cookie', b'\x00' * 8),
            ('next_payload', b'\xff'),  # Invalid payload type
            ('protocol_version', b'\x20'),  # ISAKMP v2.0
            ('exchange_type', b'\x22'),  # IKE_SA_INIT
            ('flags', b'\x00'),
            ('message_id', b'\x00' * 4),
            ('length', b'\x00\x00\x00\x1c'),  # 28 bytes
        ])
        header_bytes = b''.join(header_dict.values())
        with self.assertRaises(InvalidValue):
            IsakmpMessage.parse_exact_size(header_bytes)

    def test_unsupported_payload_type_ikev2(self):
        header_dict = collections.OrderedDict([
            ('initiator_cookie', b'\x00' * 8),
            ('responder_cookie', b'\x00' * 8),
            ('next_payload', bytes([Ikev2PayloadType.IDI.value.code])),  # valid enum, but no parser class registered
            ('protocol_version', b'\x20'),  # ISAKMP v2.0
            ('exchange_type', b'\x22'),  # IKE_SA_INIT
            ('flags', b'\x00'),
            ('message_id', b'\x00' * 4),
            ('length', b'\x00\x00\x00\x1d'),  # 29 bytes (28 header + 1 dummy payload byte)
        ])
        header_bytes = b''.join(header_dict.values()) + b'\x00'
        with self.assertRaises(InvalidType):
            IsakmpMessage.parse_exact_size(header_bytes)

    def test_get_payload_by_type_ikev2(self):
        nonce_payload = Ikev2PayloadNonce(
            flags=set(),
            nonce_data=b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'
        )
        ke_payload = Ikev2PayloadKeyExchange(
            flags=set(),
            dh_group=Ikev2DiffieHellmanGroup.MODP_GROUP_2048_BIT,
            key_exchange_data=b'\x00\x01\x02\x03\x04\x05\x06\x07'
        )
        message = IsakmpMessage(
            version=IsakmpProtocolVersion(IsakmpVersion.V2, 0),
            initiator_spi=0,
            responder_spi=0,
            exchange_type=Ikev2ExchangeType.IKE_SA_INIT,
            flags=set(),
            message_id=0,
            payloads=[nonce_payload, ke_payload]
        )

        found_nonce = message.get_payload_by_type(Ikev2PayloadType.NONCE)
        self.assertEqual(found_nonce, nonce_payload)
        self.assertEqual(found_nonce.get_payload_type(), Ikev2PayloadType.NONCE)

        found_ke = message.get_payload_by_type(Ikev2PayloadType.KE)
        self.assertEqual(found_ke, ke_payload)
        self.assertEqual(found_ke.get_payload_type(), Ikev2PayloadType.KE)

    def test_get_payload_by_type_ikev1(self):
        ke_payload = Ikev1PayloadKeyExchange(key_exchange_data=b'\x00\x01\x02\x03')
        nonce_payload = Ikev1PayloadNonce(nonce_data=b'\x00\x01\x02\x03\x04\x05\x06\x07\x08')
        message = IsakmpMessage(
            version=IsakmpProtocolVersion(IsakmpVersion.V1, 1),
            initiator_spi=0,
            responder_spi=0,
            exchange_type=Ikev1ExchangeType.BASE,
            flags=set(),
            message_id=0,
            payloads=[ke_payload, nonce_payload]
        )

        found_ke = message.get_payload_by_type(Ikev1PayloadType.KEY_EXCHANGE)
        self.assertEqual(found_ke, ke_payload)
        self.assertEqual(found_ke.get_payload_type(), Ikev1PayloadType.KEY_EXCHANGE)

        found_nonce = message.get_payload_by_type(Ikev1PayloadType.NONCE)
        self.assertEqual(found_nonce, nonce_payload)
        self.assertEqual(found_nonce.get_payload_type(), Ikev1PayloadType.NONCE)

    def test_get_payload_by_type_not_found(self):
        nonce_payload = Ikev2PayloadNonce(
            flags=set(),
            nonce_data=b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'
        )
        message = IsakmpMessage(
            version=IsakmpProtocolVersion(IsakmpVersion.V2, 0),
            initiator_spi=0,
            responder_spi=0,
            exchange_type=Ikev2ExchangeType.IKE_SA_INIT,
            flags=set(),
            message_id=0,
            payloads=[nonce_payload]
        )

        with self.assertRaises(KeyError) as context_manager:
            message.get_payload_by_type(Ikev2PayloadType.KE)
        self.assertEqual(context_manager.exception.args[0], Ikev2PayloadType.KE)

    def test_get_payload_by_type_empty_payloads(self):
        message = IsakmpMessage(
            version=IsakmpProtocolVersion(IsakmpVersion.V2, 0),
            initiator_spi=0,
            responder_spi=0,
            exchange_type=Ikev2ExchangeType.IKE_SA_INIT,
            flags=set(),
            message_id=0,
            payloads=[]
        )

        with self.assertRaises(KeyError) as context_manager:
            message.get_payload_by_type(Ikev2PayloadType.NONCE)
        self.assertEqual(context_manager.exception.args[0], Ikev2PayloadType.NONCE)



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import unittest


from cryptodatahub.common.exception import InvalidValue

from cryptoparser.common.exception import NotEnoughData

from cryptoparser.tls.subprotocol import TlsAlertMessage, TlsAlertLevel, TlsAlertDescription


class TestAlert(unittest.TestCase):
    def test_error(self):
        with self.assertRaisesRegex(InvalidValue, '0xff is not a valid TlsAlertLevel'):
            # pylint: disable=expression-not-assigned
            TlsAlertMessage.parse_exact_size(b'\xff\x00')

        with self.assertRaisesRegex(InvalidValue, '0xff is not a valid TlsAlertDescription'):
            # pylint: disable=expression-not-assigned
            TlsAlertMessage.parse_exact_size(b'\x01\xff')

        with self.assertRaises(NotEnoughData) as context_manager:
            # pylint: disable=expression-not-assigned
            TlsAlertMessage.parse_exact_size(b'\xff')
        self.assertGreaterEqual(context_manager.exception.bytes_needed, 1)

    def test_parse(self):
        self.assertEqual(
            TlsAlertMessage.parse_exact_size(b'\x02\x28'),
            TlsAlertMessage(TlsAlertLevel.FATAL, TlsAlertDescription.HANDSHAKE_FAILURE)
        )

    def test_compose(self):
        self.assertEqual(
            b'\x02\x28',
            TlsAlertMessage(TlsAlertLevel.FATAL, TlsAlertDescription.HANDSHAKE_FAILURE).compose()
        )



# SPDX-License-Identifier: MPL-2.0
"""ISAKMP header parser."""

import enum
import typing

import attr

from cryptodatahub.ike.algorithm import Ikev1PayloadType, Ikev2PayloadType, Ikev1ExchangeType, Ikev2ExchangeType

from cryptoparser.common.parse import ParsableBase, ParserBinary, ComposerBinary
from cryptoparser.common.exception import NotEnoughData, InvalidType

from cryptoparser.ike.version import IsakmpProtocolVersion, IsakmpVersion
from cryptoparser.ike.ikev1 import (
    Ikev1PayloadBase,
    IKEV1_PAYLOAD_CLASSES_BY_TYPE,
)
from cryptoparser.ike.ikev2 import (
    Ikev2PayloadBase,
    IKEV2_PAYLOAD_CLASSES_BY_TYPE,
)


class IsakmpFlags(enum.IntFlag):
    """ISAKMP flags."""
    ENCRYPTION = 1 << 0
    COMMIT = 1 << 1
    AUTHENTICATION_ONLY = 1 << 2
    INITIATOR = 1 << 3
    VERSION = 1 << 4
    RESPONSE = 1 << 5


@attr.s
class IsakmpMessage(ParsableBase):
    """ISAKMP message parser.

    .. code-block:: text

                        1                   2                   3
         0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |                       IKE SA Initiator's SPI                  |
        |                                                               |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |                       IKE SA Responder's SPI                  |
        |                                                               |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |  Next Payload | MjVer | MnVer | Exchange Type |     Flags     |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |                          Message ID                           |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |                            Length                             |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |                       Payload Data (variable)                 |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    :ivar initiator_spi: Initiator SPI
    :ivar responder_spi: Responder SPI
    :ivar exchange_type: Exchange Type
    :ivar flags: Flags
    :ivar message_id: Message ID
    :ivar payloads: Payloads
    :ivar version: Version
    """
    HEADER_SIZE = 28

    version: IsakmpProtocolVersion = attr.ib(validator=attr.validators.instance_of(IsakmpProtocolVersion))
    initiator_spi: int = attr.ib(validator=attr.validators.and_(
        attr.validators.instance_of(int),
        attr.validators.ge(0),
        attr.validators.lt(2**64)
    ))
    responder_spi: int = attr.ib(validator=attr.validators.and_(
        attr.validators.instance_of(int),
        attr.validators.ge(0),
        attr.validators.lt(2**64)))
    exchange_type: typing.Union[Ikev1ExchangeType, Ikev2ExchangeType] = attr.ib(
        validator=attr.validators.instance_of((Ikev1ExchangeType, Ikev2ExchangeType))
    )
    flags: typing.List[IsakmpFlags] = attr.ib(validator=attr.validators.deep_iterable(
        member_validator=attr.validators.instance_of(IsakmpFlags)
    ))
    message_id: int = attr.ib(validator=attr.validators.and_(
        attr.validators.instance_of(int),
        attr.validators.ge(0),
        attr.validators.lt(2**32)
    ))
    payloads: typing.List[typing.Union[Ikev1PayloadBase, Ikev2PayloadBase]] = attr.ib(
        validator=attr.validators.deep_iterable(
            member_validator=attr.validators.instance_of((Ikev1PayloadBase, Ikev2PayloadBase))
        )
    )

    def get_payload_by_type(
        self, payload_type: typing.Union[Ikev1PayloadType, Ikev2PayloadType]
    ) -> typing.Union[Ikev1PayloadBase, Ikev2PayloadBase]:
        for payload in self.payloads:
            if payload.get_payload_type() == payload_type:
                return payload

        raise KeyError(payload_type)

    @classmethod
    def _parse(cls, parsable):
        if len(parsable) < cls.HEADER_SIZE:
            raise NotEnoughData(cls.HEADER_SIZE - len(parsable))

        parser = ParserBinary(parsable)
        parser.parse_numeric('initiator_spi', 8)
        parser.parse_numeric('responder_spi', 8)
        parser.parse_raw('next_payload', 1)
        parser.parse_parsable('version', IsakmpProtocolVersion)
        if parser['version'].major == IsakmpVersion.V1:
            next_payload_type = Ikev1PayloadType
            parser.parse_numeric_enum_coded('exchange_type', Ikev1ExchangeType)
        elif parser['version'].major == IsakmpVersion.V2:
            next_payload_type = Ikev2PayloadType
            parser.parse_numeric_enum_coded('exchange_type', Ikev2ExchangeType)
        else:
            raise NotImplementedError(parser['version'])

        next_payload_parser = ParserBinary(parser['next_payload'])
        next_payload_parser.parse_numeric_enum_coded('value', next_payload_type)
        next_payload = next_payload_parser['value']

        parser.parse_numeric_flags('flags', 1, IsakmpFlags)
        parser.parse_numeric('message_id', 4)
        parser.parse_numeric('length', 4)

        version = parser['version']
        if version.major == IsakmpVersion.V1:
            payload_classes = IKEV1_PAYLOAD_CLASSES_BY_TYPE
            payload_none = Ikev1PayloadType.NONE
        elif version.major == IsakmpVersion.V2:
            payload_classes = IKEV2_PAYLOAD_CLASSES_BY_TYPE
            payload_none = Ikev2PayloadType.NONE
        else:
            raise NotImplementedError(version)

        payloads = []
        parser_payload = ParserBinary(parsable[parser.parsed_length:parser['length']])
        while parser_payload.unparsed_length > 0 and next_payload != payload_none:
            if next_payload not in payload_classes:
                raise InvalidType(next_payload)
            payload_class = payload_classes[next_payload]
            parser_payload.parse_parsable('payload', payload_class)

            payload = parser_payload['payload']
            payloads.append(payload)

            next_payload = payload.next_payload

        return cls(
            initiator_spi=parser['initiator_spi'],
            responder_spi=parser['responder_spi'],
            version=parser['version'],
            exchange_type=parser['exchange_type'],
            flags=parser['flags'],
            message_id=parser['message_id'],
            payloads=payloads,
        ), parser.parsed_length + parser_payload.parsed_length

    def compose(self):
        header_composer = ComposerBinary()

        header_composer.compose_numeric(self.initiator_spi, 8)
        header_composer.compose_numeric(self.responder_spi, 8)

        if self.version.major == IsakmpVersion.V1:
            payload_none = Ikev1PayloadType.NONE
            payload_type_security_association = (
                self.payloads[0].get_payload_type()
                if self.payloads
                else payload_none
            )
        elif self.version.major == IsakmpVersion.V2:
            payload_none = Ikev2PayloadType.NONE
            # First payload can be a cookie notification payload
            payload_type_security_association = self.payloads[0].get_payload_type()
        else:
            raise NotImplementedError(self.version)

        payload_composer = ComposerBinary()
        for i, payload in enumerate(self.payloads):
            payload.next_payload = (
                payload_none
                if i == len(self.payloads) - 1
                else self.payloads[i + 1].get_payload_type()
            )
            payload_composer.compose_parsable(payload)

        header_composer.compose_numeric_enum_coded(payload_type_security_association if self.payloads else payload_none)
        header_composer.compose_parsable(self.version)
        header_composer.compose_numeric_enum_coded(self.exchange_type)
        header_composer.compose_numeric_flags(self.flags, 1)
        header_composer.compose_numeric(self.message_id, 4)
        header_composer.compose_numeric(self.HEADER_SIZE + payload_composer.composed_length, 4)

        return header_composer.composed_bytes + payload_composer.composed_bytes



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import importlib.metadata

metadata = importlib.metadata.metadata('cryptoparser')

__title__ = metadata['Name']
__technical_name__ = __title__.lower()
__version__ = metadata['Version']
__description__ = metadata['Summary']
__author__ = metadata['Author']
__author_email__ = metadata['Author-email']
__url__ = 'https://gitlab.com/coroner/' + __technical_name__
__license__ = metadata.get('License-Expression') or metadata['License']
